"""triangulation.py: functions for finding Delaunay triangulation and Voronoi graph from a set of points."""

import logging

log = logging.getLogger(__name__)

from typing import Tuple
import math as maths
import numpy as np
from scipy.spatial import Delaunay  # type: ignore
from numba import njit  # type: ignore
from typing import Optional

import resqpy.crs as rqc
import resqpy.lines as rql
import resqpy.model as rq
import resqpy.olio.intersection as meet
import resqpy.olio.vector_utilities as vec

# _ccw_t() no longer needed: triangle vertices maintained in anti-clockwise order throughout
# def _ccw_t(p, t):   # puts triangle vertex indices into anti-clockwise order, in situ
#    if vec.clockwise(p[t[0]], p[t[1]], p[t[2]]) > 0.0:
#       t[1], t[2] = t[2], t[1]


def _dt_scipy(points: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Calculates the Delaunay triangulation for an array of points and the convex hull indices.
    
    arguments:
        points (np.ndarray): coordinates of the points to triangulate; array has shape
            (npoints, ndim)

    returns:
        (tuple): tuple containing:

            simplices (np.ndarray): indices of the points forming the triangulation simplices; array
                has shape (nsimplex, ndim+1)
            convex_hull_indices (np.ndarray): indices of the points forming the convex hull; array
                has shape (nhull,)

    note:
        the triangulation is carried out on the points as projected onto the xy plane
    """
    delaunay = Delaunay(points[..., :2])
    simplices = delaunay.simplices
    convex_hull_indices = np.unique(delaunay.convex_hull)
    return simplices, convex_hull_indices


def _dt_simple(po, plot_fn = None, progress_fn = None, container_size_factor = None):
    # returns Delauney triangulation of po and list of hull point indices, using a simple algorithm

    def flip(ei):
        nonlocal fm, e, t, te, p, nt, p_i, ne
        if fm[ei]:
            return  # this edge has already been flipped since last point insertion
        t0, te0 = e[ei, 0]
        t1, te1 = e[ei, 1]
        if t0 < 0 or t1 < 0:
            return  # no triangle on other side
        t0n = te0 - 1
        if t0n < 0:
            t0n = 2
        t1n = te1 - 1
        if t1n < 0:
            t1n = 2
        if (not vec.in_circumcircle(p[t[t0, 0]], p[t[t0, 1]], p[t[t0, 2]], p[t[t1, t1n]]) and
                not vec.in_circumcircle(p[t[t1, 0]], p[t[t1, 1]], p[t[t1, 2]], p[t[t0, t0n]])):
            return  # not sure both are needed
        # flip needed
        ft0 = (t[t0, te0], t[t1, t1n], t[t0, t0n])
        ft1 = (t[t1, te1], t[t0, t0n], t[t1, t1n])
        e[ei, :, 1] = 1
        fte0 = (te[t1, 3 - (te1 + t1n)], ei, te[t0, t0n])
        fte1 = (te[t0, 3 - (te0 + t0n)], ei, te[t1, t1n])
        if e[fte0[0], 0, 0] == t1:
            e[fte0[0], 0] = (t0, 0)
        elif e[fte0[0], 1, 0] == t1:
            e[fte0[0], 1] = (t0, 0)
        else:
            raise Exception('edge breakdown')
        if e[fte1[0], 0, 0] == t0:
            e[fte1[0], 0] = (t1, 0)
        elif e[fte1[0], 1, 0] == t0:
            e[fte1[0], 1] = (t1, 0)
        else:
            raise Exception('edge breakdown')
        if e[fte0[2], 0, 0] == t0:
            e[fte0[2], 0, 1] = 2
        elif e[fte0[2], 1, 0] == t0:
            e[fte0[2], 1, 1] = 2
        else:
            raise Exception('edge breakdown')
        if e[fte1[2], 0, 0] == t1:
            e[fte1[2], 0, 1] = 2
        elif e[fte1[2], 1, 0] == t1:
            e[fte1[2], 1, 1] = 2
        else:
            raise Exception('edge breakdown')
        t[t0] = ft0
        t[t1] = ft1
        te[t0] = fte0
        te[t1] = fte1
        fm[ei] = True
        # debug plot here
        if plot_fn is not None:
            plot_fn(p, t[:nt])
        # recursively flip, not sure all these are needed
        flip(fte0[0])
        flip(fte0[2])
        flip(fte1[0])
        flip(fte1[2])

    n_p = len(po)
    if n_p < 3:
        return None, None  # not enough points
    elif n_p == 3:
        return np.array([0, 1, 2], dtype = int).reshape((1, 3)), np.array([0, 1, 2], dtype = int)

    if progress_fn is not None:
        progress_fn(0.0)

    min_xy = np.min(po[:, :2], axis = 0)
    max_xy = np.max(po[:, :2], axis = 0)
    dxy = max_xy - min_xy
    assert dxy[0] > 0.0 and dxy[1] > 0.0, 'points lie in straight line or are conincident'
    p = np.empty((n_p + 3, 2))
    p[:-3] = po[:, :2]
    # add 3 points sure of containing all po
    csf = 100.0 if container_size_factor is None else max(1.0, container_size_factor)
    p[-3] = (min_xy[0] - 0.8 * csf * dxy[0], min_xy[1] - 0.1 * csf * dxy[1])
    p[-2] = (max_xy[0] + 0.8 * csf * dxy[0], min_xy[1] - 0.1 * csf * dxy[1])
    p[-1] = (min_xy[0] + 0.5 * csf * dxy[0], max_xy[1] + 0.8 * csf * dxy[1])

    # triangle vertex indices
    t = np.empty((2 * n_p + 2, 3), dtype = int)  # empty space for triangle vertex indices
    t[0] = (n_p, n_p + 1, n_p + 2)  # initial set of one containing triangle
    nt = 1  # number of triangles so far populated

    # edges: list of indices of triangles and edge within triangle; -1 indicates no triangle using edge
    e = np.full((3 * n_p + 6, 2, 2), fill_value = -1, dtype = int)
    e[:3, 0, 0] = 0
    for edge in range(3):
        e[edge, 0, 1] = edge
    ne = 3  # number of edges so far in use

    # edge indices (in e) for each triangle, first axis indexed in sync with t
    te = np.empty((2 * n_p + 2, 3), dtype = int)  # empty space for triangle edge indices
    te[0] = (0, 1, 2)

    # mask tracking which edges have been flipped
    fm = np.zeros((3 * n_p + 6), dtype = bool)

    # debug plot here
    #   if plot_fn: plot_fn(p, t[:nt])

    progress_period = max(n_p // 100, 1)
    progress_count = progress_period

    for p_i in range(n_p):  # index of next point to consider

        if progress_fn is not None and progress_count <= 0:
            progress_fn(float(p_i) / float(n_p))
            progress_count = progress_period

        # find triangle that contains this point
        f_t = None
        for ti in range(nt):
            if vec.in_triangle_edged(p[t[ti, 0]], p[t[ti, 1]], p[t[ti, 2]], p[p_i]):
                f_t = ti
                break
        assert f_t is not None, 'failed to find triangle containing point'

        # take copy of edge indices for containing triangle
        e0, e1, e2 = te[f_t]

        # split containing triangle into 3, using new point as common vertex
        t[nt] = (t[f_t, 1], t[f_t, 2], p_i)
        t[nt + 1] = (t[f_t, 2], t[f_t, 0], p_i)
        t[f_t, 2] = p_i
        # add 3 new edges and update te
        e[ne, 0] = (f_t, 2)
        e[ne, 1] = (nt + 1, 1)
        e[ne + 1, 0] = (f_t, 1)
        e[ne + 1, 1] = (nt, 2)
        e[ne + 2, 0] = (nt, 1)
        e[ne + 2, 1] = (nt + 1, 2)
        if e[e1, 0, 0] == f_t:
            e[e1, 0] = (nt, 0)
        elif e[e1, 1, 0] == f_t:
            e[e1, 1] = (nt, 0)
        else:
            raise Exception('edge breakdown')
        if e[e2, 0, 0] == f_t:
            e[e2, 0] = (nt + 1, 0)
        elif e[e2, 1, 0] == f_t:
            e[e2, 1] = (nt + 1, 0)
        else:
            raise Exception('edge breakdown')
        te[nt] = (e1, ne + 2, ne + 1)
        te[nt + 1] = (e2, ne, ne + 2)
        te[f_t, 1] = ne + 1
        te[f_t, 2] = ne

        nt += 2
        ne += 3

        # now recursively try flipping sides
        fm[:] = False

        # debug plot here, with new point, before flipping
        if plot_fn is not None:
            plot_fn(p, t[:nt])

        flip(e0)
        flip(e1)
        flip(e2)

        progress_count -= 1

    # remove any triangles using invented container vertices
    tri_set = t[np.where(np.all(t[:nt] < n_p, axis = 1))]
    if plot_fn is not None:
        plot_fn(p, tri_set)

    external_t = t[np.where(np.any(t[:nt] >= n_p, axis = 1))]
    external_pi = np.unique(external_t)[:-3]  # discard 3 invented container vertices

    if progress_fn is not None:
        progress_fn(1.0)

    return tri_set, external_pi


def dt(p, algorithm = "scipy", plot_fn = None, progress_fn = None, container_size_factor = 100.0, return_hull = False):
    """Returns the Delauney Triangulation of 2D point set p.

    arguments:
       p (numpy float array of shape (N, 2): the x,y coordinates of the points
       algorithm (string, optional): selects which algorithm to use; current options: ['simple', 'scipy'];
          if None, the current best algorithm is selected
       plot_fn (function of form f(p, t), optional): if present, this function is called each time the
          algorithm feels it is worth refreshing a plot of the progress; p is a copy of the point set,
          depending on the algorithm with 3 extra points added to form an enveloping triangle
       progress_fn (function of form f(x), optional): if present, this function is called at regulat
          intervals by the algorithm, passing increasing values in the range 0.0 to 1.0 as x
       container_size_factor (float, default 100.0): the larger this number, the more likely the
          resulting triangulation is to be convex; reduce to 1.0 to allow slight concavities
       return_hull (boolean, default False): if True, a pair is returned with the second item being
          a clockwise ordered list of indices into p identifying the points on the boundary of the
          returned triangulation

    returns:
       numpy int array of shape (M, 3) - being the indices into the first axis of p of the 3 points
          per triangle in the Delauney Triangulation - and if return_hull is True, another int array
          of shape (B,) - being indices into p of the clockwise ordered points on the boundary of
          the triangulation

    notes:
       the plot_fn, progress_fn and container_size_factor arguments are only used by the 'simple' algorithm;
       if points p are 3D, the projection onto the xy plane is used for the triangulation
    """
    assert p.ndim == 2 and p.shape[1] >= 2, 'bad points shape for 2D Delauney Triangulation'

    if not algorithm:
        algorithm = 'scipy'

    if algorithm == 'scipy':
        tri, boundary = _dt_scipy(p)
    elif algorithm == 'simple':
        tri, boundary = _dt_simple(p,
                                   plot_fn = plot_fn,
                                   progress_fn = progress_fn,
                                   container_size_factor = container_size_factor)
    else:
        raise Exception(f'unrecognised Delauney Triangulation algorithm name: {algorithm}')

    assert tri.ndim == 2 and tri.shape[1] == 3

    if return_hull:
        return tri, vec.clockwise_sorted_indices(p, boundary)
    else:
        return tri


def ccc(p1, p2, p3):
    """Returns the centre of the circumcircle of the three points in the xy plane."""

    # two edges as vectors
    v12 = p2 - p1
    v13 = p3 - p1
    # midpoints of the two edges
    m12 = 0.5 * (p1 + p2)
    m13 = 0.5 * (p1 + p3)
    # pairs of points defining two edge normals passing through the midpoints
    o12 = m12.copy()
    o12[0] += v12[1]
    o12[1] -= v12[0]
    o13 = m13.copy()
    o13[0] += v13[1]
    o13[1] -= v13[0]
    return meet.line_line_intersect(m12[0], m12[1], o12[0], o12[1], m13[0], m13[1], o13[0], o13[1])


def voronoi(p, t, b, aoi: rql.Polyline):
    """Returns dual Voronoi diagram for a Delauney triangulation.

    arguments:
       p (numpy float array of shape (N, 2)): seed points used in the Delauney triangulation
       t (numpy int array of shape (M, 3)): the Delauney triangulation of p as returned by dt()
       b (numpy int array of shape (B,)): clockwise sorted list of indices into p of the boundary
          points of the triangulation t
       aoi (lines.Polyline): area of interest; a closed clockwise polyline that must strictly contain
          all p (no points exactly on or outside the polyline)

    returns:
       c, v where: c is a numpy float array of shape (M+E, 2) being the circumcircle centres of
       the M triangles and E boundary points from the aoi polygon line; and v is a list of
       N Voronoi cell lists of clockwise ints, each int being an index into c

    notes:
       the aoi polyline forms the outer boundary for the Voronoi polygons for points on the
       outer edge of the triangulation; all points p must lie strictly within the aoi, which
       must be convex; the triangulation t, of points p, must also have a convex hull; note
       that the dt() function can produce a triangulation with slight concavities on the hull,
       especially for smaller values of its container_size_factor argument
    """

    # this code assumes that the Voronoi polygon for a seed point visits the circumcentres of
    # all the triangles that make use of the point – currently understood to be always the case
    # for a Delauney triangulation

    def __aoi_intervening_nodes(aoi_count, c_count, seg_a, seg_c):
        nodes = []
        seg = seg_a
        while seg != seg_c:
            seg = (seg + 1) % aoi_count
            nodes.append(c_count + seg)
        return nodes

    def __shorter_sides_p_i(p3):
        max_length = -1.0
        max_i = None
        for i in range(3):
            opp_length = vec.naive_length(p3[i - 1] - p3[i - 2])
            if opp_length > max_length:
                max_length = opp_length
                max_i = i
        return max_i

    def __azi_between(a, c, t):
        if c < a:
            c += 360.0
        return (a <= t <= c) or (a <= t + 360.0 <= c)

    def __seg_for_ci(ci):  # returns hull segment for a boundary index
        nonlocal ca_count, cah_count, caho_count, cahon_count, wing_hull_segments
        if ci < ca_count:
            return None
        if ci < cah_count:  # hull edge intersection
            return ci - ca_count
        if ci < caho_count:  # wings
            oi, wing = divmod(ci - cah_count, 2)
            return wing_hull_segments[oi, wing]
        if ci < cahon_count:  # virtual centre for hull edge
            return ci - caho_count
        # else virtual centre for hull point; arbitrarily pick clockwise segment
        return ci - cahon_count

    def __intervening_aoi_indices(aoi_count, aoi_intersect_segments, c_count, ca_count, cah_count, ci_for_p,
                                  out_pair_intersect_segments):
        # build list of intervening aoi boundary point indices and append to list
        aoi_nodes = []
        r = [0] if len(ci_for_p) == 2 else range(len(ci_for_p))
        just_done_pair = False
        for cii in r:
            cip = ci_for_p[cii]
            ci = ci_for_p[(cii + 1) % len(ci_for_p)]
            if cip >= c_count and ci >= c_count and not just_done_pair:
                # identify aoi segments
                if cip < cah_count:
                    aoi_seg_a = aoi_intersect_segments[cip - ca_count]
                else:
                    aoi_seg_a = out_pair_intersect_segments[divmod(cip - cah_count, 2)]
                if ci < cah_count:
                    aoi_seg_c = aoi_intersect_segments[ci - ca_count]
                else:
                    aoi_seg_c = out_pair_intersect_segments[divmod(ci - cah_count, 2)]
                aoi_nodes += __aoi_intervening_nodes(aoi_count, c_count, aoi_seg_a, aoi_seg_c)
                just_done_pair = True
            else:
                just_done_pair = False
        return aoi_nodes

    def __ci_non_hull(b_i, c, c_count, ci_for_p, p, p_i):
        trimmed_ci = []
        cii = 0
        finish_at = len(ci_for_p)
        while cii < finish_at:
            if ci_for_p[cii] < c_count:
                trimmed_ci.append(ci_for_p[cii])
                cii += 1
                continue
            start_cii = cii
            cii_seg = __seg_for_ci(ci_for_p[cii])
            while cii == 0 and ci_for_p[start_cii - 1] >= c_count and __seg_for_ci(ci_for_p[start_cii - 1]) == cii_seg:
                start_cii -= 1
            start_cii = start_cii % len(ci_for_p)
            end_cii = cii + 1
            while end_cii < len(ci_for_p) and ci_for_p[end_cii] >= c_count and __seg_for_ci(
                    ci_for_p[end_cii]) == cii_seg:
                end_cii += 1
            end_cii -= 1  # unpythonesque: end element included in scan
            if end_cii == start_cii:
                trimmed_ci.append(ci_for_p[cii])
                cii += 1
                continue
            if end_cii < start_cii:
                finish_at = start_cii
            if end_cii == (start_cii + 1) % len(ci_for_p):
                trimmed_ci.append(ci_for_p[start_cii])
                trimmed_ci.append(ci_for_p[end_cii])
                cii = end_cii + 1
                continue
            start_azi = vec.azimuth(c[ci_for_p[start_cii]] - p[p_i, :2])
            end_azi = vec.azimuth(c[ci_for_p[end_cii]] - p[p_i, :2])
            scan_cii = start_cii
            while True:
                if scan_cii == start_cii:
                    trimmed_ci.append(ci_for_p[scan_cii])
                elif scan_cii == end_cii:
                    trimmed_ci.append(ci_for_p[scan_cii])
                    break
                else:
                    # if point is around a hull corner from previous, then include (?)
                    next_cii = (scan_cii + 1) % len(ci_for_p)
                    if b_i is None and __seg_for_ci(ci_for_p[scan_cii]) != __seg_for_ci(ci_for_p[next_cii]):
                        trimmed_ci.append(ci_for_p[scan_cii])
                        trimmed_ci.append(ci_for_p[next_cii])
                        scan_cii = next_cii
                    elif b_i is None and __seg_for_ci(trimmed_ci[-1]) != __seg_for_ci(ci_for_p[scan_cii]):
                        trimmed_ci.append(ci_for_p[scan_cii])
                    else:
                        azi = vec.azimuth(c[ci_for_p[scan_cii]] - p[p_i, :2])
                        if __azi_between(start_azi, end_azi, azi):
                            trimmed_ci.append(ci_for_p[scan_cii])
                if scan_cii == end_cii:
                    break
                scan_cii = (scan_cii + 1) % len(ci_for_p)
            cii = end_cii + 1
        return trimmed_ci

    def __closest_to_seed(c, c_count, ci_for_p, hull_node_azi, p, p_i):
        best_a_i = None
        best_c_i = None
        best_a_azi = -181.0
        best_c_azi = 181.0
        for cii, val in enumerate(ci_for_p):
            if val < c_count:
                continue
            azi = vec.azimuth(c[val] - p[p_i, :2]) - hull_node_azi
            if azi > 180.0:
                azi -= 360.0
            elif azi < -180.0:
                azi += 360.0
            if 0.0 > azi > best_a_azi:
                best_a_azi = azi
                best_a_i = cii
            elif 0.0 <= azi < best_c_azi:
                best_c_azi = azi
                best_c_i = cii
        assert best_a_i is not None and best_c_i is not None
        trimmed_ci = []
        for cii, val in enumerate(ci_for_p):
            if val < c_count or cii == best_a_i or cii == best_c_i:
                trimmed_ci.append(val)
        return trimmed_ci

    def __ci_replace(c_count, ca_count, cah_count, caho_count, cahon_count, ci_for_p, p, p_i, t, tc_outwith_aoi):
        # where circumcirle (or virtual) centre is outwith aoi, replace with a point on aoi boundary
        # virtual centres related to hull points (not hull edges) can be discarded
        trimmed_ci = []
        for ci in ci_for_p:
            if ci < c_count:  # genuine triangle
                if ci in tc_outwith_aoi:  # replace with one or two wing normal intersection points
                    oi = tc_outwith_aoi.index(ci)
                    wing_i = cah_count + 2 * oi
                    shorter_t_i = __shorter_sides_p_i(p[t[ci]])
                    if t[ci, shorter_t_i] == p_i:
                        trimmed_ci += [wing_i, wing_i + 1]
                    elif t[ci, shorter_t_i - 1] == p_i:
                        trimmed_ci.append(wing_i)
                    else:
                        trimmed_ci.append(wing_i + 1)
                else:
                    trimmed_ci.append(ci)
            elif ci < cahon_count:
                # extended virtual centre for a hull edge (discard hull point virtual centres)
                # replace with index for intersection point on aoi boundary
                trimmed_ci.append(ca_count + ci - caho_count)
        return trimmed_ci

    def __veroni_cells(aoi_count, aoi_intersect_segments, b, c, c_count, ca_count, cah_count, caho_count, cahon_count,
                       hull_count, out_pair_intersect_segments, p, t, tc_outwith_aoi):
        # list of voronoi cells (each a numpy list of node indices into c extended with aoi points etc)
        v = []
        # for each seed point build the voronoi cell
        for p_i in range(len(p)):

            # find triangles making use of that point
            ci_for_p = list(np.where(t == p_i)[0])

            # if seed point is on hull boundary, introduce three extended virtual centres
            b_i = None
            if p_i in b:
                b_i = np.where(b == p_i)[0][0]  # index into hull coordinates
                p_b_i = (b_i - 1) % hull_count  # predecessor, ie. anti-clockwise boundary point
                ci_for_p += [caho_count + p_b_i, cahon_count + b_i, caho_count + b_i]

            # find azimuths of vectors from seed point to circumcircle centres (and virtual centres)
            azi = [vec.azimuth(centre - p[p_i, :2]) for centre in c[ci_for_p, :2]]
            # if this is a hull seed point, make a note of azimuth to virtual centre
            hull_node_azi = None if b_i is None else azi[-2]
            # sort triangle indices for seed point into clockwise order of circumcircle (and virtual) centres
            ci_for_p = [ti for (_, ti) in sorted(zip(azi, ci_for_p))]

            ci_for_p = __ci_replace(c_count, ca_count, cah_count, caho_count, cahon_count, ci_for_p, p, p_i, t,
                                    tc_outwith_aoi)

            # if this is a hull seed point, classify aoi boundary points into anti-clockwise or clockwise, and find
            # closest to seed
            if b_i is not None:
                ci_for_p = __closest_to_seed(c, c_count, ci_for_p, hull_node_azi, p, p_i)

            # for sequences on aoi boundary, just keep those between the first and last (?)

            #      elif any([ci < c_count for ci in ci_for_p]):
            else:
                ci_for_p = __ci_non_hull(b_i, c, c_count, ci_for_p, p, p_i)

            # reverse points if needed for pair of aoi points only
            assert len(ci_for_p) >= 2
            if len(ci_for_p) == 2:
                seg_0 = __seg_for_ci(ci_for_p[0])
                seg_1 = __seg_for_ci(ci_for_p[1])
                if seg_0 is not None and seg_1 is not None and seg_0 == (seg_1 + 1) % hull_count:
                    ci_for_p.reverse()

            ci_for_p += __intervening_aoi_indices(aoi_count, aoi_intersect_segments, c_count, ca_count, cah_count,
                                                  ci_for_p, out_pair_intersect_segments)

            #  remove circumcircle centres that are outwith area of interest
            ci_for_p = np.array([ti for ti in ci_for_p if ti >= c_count or ti not in tc_outwith_aoi], dtype = int)

            # find azimuths of vectors from seed point to circumcircle centres and aoi boundary points
            azi = [vec.azimuth(centre - p[p_i, :2]) for centre in c[ci_for_p, :2]]

            # re-sort triangle indices for seed point into clockwise order of circumcircle centres and boundary points
            ordered_ci = [ti for (_, ti) in sorted(zip(azi, ci_for_p))]

            v.append(ordered_ci)
        return v

    # log.debug(f'\n\nVoronoi: nt: {len(p)}; nt: {len(t)}; hull: {len(b)}; aoi: {len(aoi.coordinates)}')
    # todo: allow aoi to be None in which case create an aoi as hull with border

    assert p.ndim == 2 and p.shape[0] > 2 and p.shape[1] >= 2
    assert t.ndim == 2 and t.shape[1] == 3
    assert b.ndim == 1 and b.shape[0] > 2
    assert len(aoi.coordinates) >= 3
    assert aoi.isclosed
    assert aoi.is_clockwise()

    # create temporary polyline for hull of triangulation
    hull = rql.Polyline(aoi.model,
                        is_closed = True,
                        set_coord = p[b],
                        set_crs = aoi.crs_uuid,
                        title = 'triangulation hull')
    hull_count = len(b)

    # check for concavities in hull
    if not hull.is_convex():
        log.warning('Delauney triangulation is not convex; Voronoi diagram construction might fail')

    # compute circumcircle centres
    c = np.zeros((t.shape[0], 2))
    for ti in range(len(t)):
        c[ti] = ccc(p[t[ti, 0]], p[t[ti, 1]], p[t[ti, 2]])
    c_count = len(c)

    # make list of triangle indices whose circumcircle centres are outwith the area of interest
    tc_outwith_aoi = [ti for ti in range(c_count) if not aoi.point_is_inside_xy(c[ti])]
    o_count = len(tc_outwith_aoi)

    # make space for combined points data needed for all voronoi cell nodes:
    # 1. circumcircle centres for triangles in delauney triangulation
    # 2. nodes defining area of interest polygon
    # 3. intersection of normals to triangulation hull edges with aoi polygon
    # 4. extra intersections for normals to other two (non-hull) triangle edges, with aoi,
    #    where circumcircle centre is outside the area of interest
    # 5. extended ccc for hull edge normals
    # 6. extended ccc for hull points
    # (5 & 6 only used during construction)
    c = np.concatenate((c, aoi.coordinates[:, :2], np.zeros(
        (hull_count, 2), dtype = float), np.zeros(
            (2 * o_count, 2), dtype = float), np.zeros((hull_count, 2),
                                                       dtype = float), np.zeros((hull_count, 2), dtype = float)))
    aoi_count = len(aoi.coordinates)
    ca_count = c_count + aoi_count
    cah_count = ca_count + hull_count
    caho_count = cah_count + 2 * o_count
    cahon_count = caho_count + hull_count
    assert cahon_count + hull_count == len(c)

    #  compute intersection points between hull edge normals and aoi polyline
    # also extended virtual centres for hull edges
    extension_scaling = 1000.0 * np.sum((np.max(aoi.coordinates, axis = 0) - np.min(aoi.coordinates, axis = 0))[:2])
    aoi_intersect_segments = np.empty((hull_count,), dtype = int)
    for ei in range(hull_count):
        # use segment midpoint and normal methods of hull to project out
        m = hull.segment_midpoint(ei)[:2]  # midpoint
        norm_vec = hull.segment_normal(ei)[:2]
        n = m + norm_vec  # point on normal
        # use first intersection method of aoi to intersect projected normal from triangulation hull
        aoi_seg, aoi_x, aoi_y = aoi.first_line_intersection(m[0], m[1], n[0], n[1], half_segment = True)
        assert aoi_seg is not None
        # inject intersection points to extension area of c and take note of aoi segment of intersection
        c[ca_count + ei] = (aoi_x, aoi_y)
        aoi_intersect_segments[ei] = aoi_seg
        # inject extended virtual circle centres for hull edges, a long way out
        c[caho_count + ei] = c[ca_count + ei] + extension_scaling * norm_vec

    # compute extended virtual centres for hull nodes
    for ei in range(hull_count):
        pei = (ei - 1) % hull_count
        vector = vec.unit_vector(hull.segment_normal(pei)[:2] + hull.segment_normal(ei)[:2])
        c[cahon_count + ei] = hull.coordinates[ei, :2] + extension_scaling * vector

    # where cicrumcircle centres are outwith aoi, compute intersections of normals of wing edges with aoi
    out_pair_intersect_segments = np.empty((o_count, 2), dtype = int)
    wing_hull_segments = np.empty((o_count, 2), dtype = int)
    for oi, ti in enumerate(tc_outwith_aoi):
        tpi = __shorter_sides_p_i(p[t[ti]])
        for wing in range(2):
            # note: triangle nodes are anticlockwise
            m = 0.5 * (p[t[ti, tpi - 1]] + p[t[ti, tpi]])[:2]  # triangle edge midpoint
            edge_v = p[t[ti, tpi]] - p[t[ti, tpi - 1]]
            n = m + np.array((-edge_v[1], edge_v[0]))  # point on perpendicular bisector of triangle edge
            o_seg, o_x, o_y = aoi.first_line_intersection(m[0], m[1], n[0], n[1], half_segment = True)
            c[cah_count + 2 * oi + wing] = (o_x, o_y)
            out_pair_intersect_segments[oi, wing] = o_seg
            wing_hull_segments[oi, wing], _, _ = hull.first_line_intersection(m[0],
                                                                              m[1],
                                                                              n[0],
                                                                              n[1],
                                                                              half_segment = True)
            tpi = (tpi + 1) % 3

    v = __veroni_cells(aoi_count, aoi_intersect_segments, b, c, c_count, ca_count, cah_count, caho_count, cahon_count,
                       hull_count, out_pair_intersect_segments, p, t, tc_outwith_aoi)

    return c[:caho_count], v


def triangulated_polygons(p, v, centres = None):
    """Returns triangulation of polygons using centres as extra points.

    arguments:
       p (2D numpy float array): points used as vertices of polygons
       v (list of list of ints): ordered indices into p for each polygon
       centres (2D numpy float array, optional): the points to use as the centre for each polygon

    returns:
       points, triangles where: points is a copy of p extended with the centre points of polygons;
       and triangles is a numpy int array of shape (N, 3) being the triangulation of points, where N is
       equal to the overall length of v

    notes:
       if no centres are provided, balanced centre points are computed for the polygons;
       the polygons must be convex (at least from the perspective of the centre points);
       the clockwise/anti-clockwise order of the triangle edges will match that of the polygon;
       the centre point is the first point in each triangle;
       the order of triangles will match the order of vertices in a flattened view of list v;
       p and centres may have a shape of 2 or 3 in the second dimension (xy or xyz data);
       p & v could be the values (c, v) returned by the voronoi() function, in which case the
       original seed points p passed into voronoi() can be passed as centres here
    """

    assert p.ndim == 2 and p.shape[1] in [2, 3]
    assert len(v) > 0
    if centres is not None:
        assert centres.ndim == 2
        assert len(centres) == len(v)
        assert centres.shape[1] == p.shape[1]

    model = rq.Model(create_basics = True)  # temporary object for handling Polylines
    crs = rqc.Crs(model)

    points = np.zeros((len(p) + len(v), p.shape[1]))  # polygon nodes, to be extended with centre points
    points[:len(p)] = p
    if centres is not None:
        points[len(p):] = centres

    t_count = sum([len(x) for x in v])
    triangles = np.empty((t_count, 3), dtype = int)
    t_index = 0

    for cell, poly_vertices in enumerate(v):
        # add polygon centre points to points array, if not already provided
        centre_i = len(p) + cell
        if centres is None:
            polygon = rql.Polyline(model,
                                   set_coord = p[np.array(poly_vertices, dtype = int)],
                                   is_closed = True,
                                   set_crs = crs.uuid,
                                   title = 'v cell')
            poly_centre = polygon.balanced_centre()
            points[centre_i] = poly_centre[:p.shape[1]]
        # and populate triangles for this polygon
        for ti in range(len(poly_vertices)):
            triangles[t_index] = (centre_i, poly_vertices[ti], poly_vertices[(ti + 1) % len(poly_vertices)])
            t_index += 1
    assert t_index == t_count

    return points, triangles


def reorient(points, rough = True, max_dip = None, use_linalg = True, sample = 500):
    """Returns a reoriented copy of a set of points, such that z axis is approximate normal to average plane of points.

    arguments:
       points (numpy float array of shape (..., 3)): the points to be reoriented
       rough (bool, default True): if True, the resulting orientation will be within around 10 degrees of the optimum;
          if False, that reduces to around 2.5 degrees of the optimum; iugnored if use_linalg is True
       max_dip (float, optional): if present, the reorientation of perspective off vertical is
          limited to this angle in degrees
       use_linalg (bool, default True): if True, the numpy linear algebra svd function is used and rough is ignored
       sample (int, default 500): downsample points to this number for the purposes of determining normal vector

    returns:
       numpy float array of the same shape as points, numpy xyz vector, numpy 3x3 matrix;
       the array being a copy of points rotated in 3D space to minimise the z range;
       the vector is a normal vector to the original points;
       the matrix is rotation matrix used to transform the original points to the reoriented points

    notes:
       the original points array is not modified by this function;
       implicit xy & z units for points are assumed to be the same;
       the function may typically be called prior to the Delauney triangulation, which uses an xy projection to
       determine the triangulation;
       the numpy linear algebra option seems to be memory intensive, not recommended;
       downsampling will occur (for normal vector determination) when the number of points exceeds double that
       given in the sample argument; set sample to None to use all points for normal vector determination
    """

    def best_angles(points, mid_x, mid_y, steps, d_theta):
        best_range = None
        best_x_rotation = None
        best_y_rotation = None
        half_steps = float(steps - 1) / 2.0
        for xi in range(steps):
            x_degrees = mid_x + (float(xi) - half_steps) * d_theta
            for yi in range(steps):
                y_degrees = mid_y + (float(yi) - half_steps) * d_theta
                rotation_m = vec.rotation_3d_matrix((x_degrees, 0.0, y_degrees))
                p = points.copy()
                rotated_p = vec.rotate_array(rotation_m, p)
                z_r = np.nanmax(rotated_p[..., 2]) - np.nanmin(rotated_p[..., 2])
                if best_range is None or z_r < best_range:
                    best_range = z_r
                    best_x_rotation = x_degrees
                    best_y_rotation = y_degrees
        return (best_x_rotation, best_y_rotation)

    def linalg_normal_vector(p):
        assert p.ndim >= 2 and p.shape[-1] == 3
        p = p.reshape((-1, 3))
        centre = p.sum(axis = 0) / p.shape[0]
        _, _, vh = np.linalg.svd(p - centre)
        # unit normal vector
        return vh[2, :]

    assert points.ndim >= 2 and points.shape[-1] == 3

    if sample is not None and len(points) > 2 * sample:
        step = len(points) // sample
        p = points[::step, :]
    else:
        p = points

    if use_linalg:
        normal_vector = linalg_normal_vector(p)
        incl = vec.inclination(normal_vector)
        if incl == 0.0:
            rotation_m = vec.no_rotation_matrix()
        else:
            azi = vec.azimuth(normal_vector)
            rotation_m = vec.tilt_3d_matrix(azi, incl)
    else:
        # coarse iteration trying a few different angles
        best_x_rotation, best_y_rotation = best_angles(p, 0.0, 0.0, 7, 30.0)

        # finer iteration searching around the best coarse rotation
        best_x_rotation, best_y_rotation = best_angles(p, best_x_rotation, best_y_rotation, 5, 10.0)

        if not rough:
            # finer iteration searching around the best coarse rotation
            best_x_rotation, best_y_rotation = best_angles(p, best_x_rotation, best_y_rotation, 7, 2.5)

        rotation_m = vec.rotation_3d_matrix((best_x_rotation, 0.0, best_y_rotation))

        normal_vector = vec.rotate_vector(rotation_m.T, np.array((0.0, 0.0, 1.0)))

    if max_dip is not None:
        v = vec.rotate_vector(rotation_m.T, np.array((0.0, 0.0, 1.0)))
        incl = vec.inclination(v)
        if incl > max_dip:
            azi = vec.azimuth(v)
            rotation_m = vec.tilt_3d_matrix(azi, max_dip)  # TODO: check whether any reverse direction errors here
            normal_vector = vec.rotate_vector(rotation_m.T, np.array((0.0, 0.0, 1.0)))

    p = points.copy()

    return vec.rotate_array(rotation_m, p), normal_vector, rotation_m


def make_all_clockwise_xy(t, p):
    """Modifies t in situ such that each triangle is clockwise in xy plane (viewed from -ve z axis).

    note:
       assumes xyz axes are left handed; all will be made anti-clockwise in the case of right handed xyz axes
    """

    cw = (vec.clockwise_triangles(p, t, projection = 'xy') > 0.0)
    t_flip = t.copy()
    t_flip[:, 0] = t[:, 1]
    t_flip[:, 1] = t[:, 0]
    t[:] = np.where(np.expand_dims(cw, axis = 1), t, t_flip)
    return t


def surrounding_xy_ring(p, count = 12, radial_factor = 10.0, radial_distance = None, inner_ring = False):
    """Creates a set of points surrounding the point set p, in the xy plane.

    arguments:
       p (numpy float array of shape (..., 3)): xyz set of points to be surrounded
       count (int): the number of points to generate in the surrounding ring
       radial_factor (float): a distance factor roughly determining the radius of the ring relative to
          the 'radius' of the outermost points in p
       radial_distance (float): if present, the radius of the ring of points, unless radial_factor
          results in a greater distance in which case that is used
       inner_ring (bool, default False): if True, an inner ring of points, with double count, is created
          at a radius just outside that of the furthest flung original point; this improves triangulation
          of the extended point set when the original has a non-convex hull

    returns:
       numpy float array of shape (N, 3) being xyz points in surrounding ring(s); z is set constant to
       mean value of z in p; N is count if inner_ring is False, 3 * count if True
    """

    def make_ring(count, centre, radius):
        delta_theta = 2.0 * maths.pi / float(count)
        ring = np.zeros((count, 3))
        for i in range(count):
            theta = float(i) * delta_theta
            ring[i] = centre + radius * np.array([maths.cos(theta), maths.sin(theta), 0.0])
        return ring

    assert p.shape[-1] == 3
    assert radial_factor >= 1.0
    centre = np.nanmean(p.reshape((-1, 3)), axis = 0)
    p_radius_v = np.nanmax(np.abs(p.reshape((-1, 3)) - np.expand_dims(centre, axis = 0)), axis = 0)[:2]
    p_radius = maths.sqrt(np.sum(p_radius_v * p_radius_v))
    radius = p_radius * radial_factor
    if radial_distance is not None and radial_distance > radius:
        radius = radial_distance
    ring = make_ring(count, centre, radius)
    if inner_ring:
        inner_radius = p_radius * 1.1
        assert radius > inner_radius
        in_ring = make_ring(2 * count, centre, inner_radius)
        return np.concatenate((in_ring, ring), axis = 0)
    return ring


def edges(t):
    """Returns unique edges as pairs of point indices, and a count of uses of each edge.

    arguments:
        t (numpy int array of shape (N, 3)): the points indices defining a set of triangles

    returns:
        (numpy int array of shape (N, 2), numpy int array of shape (N,))
        where 2D array is list-like sorted points index pairs for unique edges
        and 1D array contains corresponding edge usage count (usually 1 or 2)

    notes:
        first entry in each pair is always the lower of the two point indices;
        for well formed surfaces, the count should everywhere be zero or one;
        the function does not attempt to detect coincident points
    """

    assert t.ndim == 2 and t.shape[1] == 3
    all_edges = np.empty((len(t), 3, 2), dtype = int)
    all_edges[:, :, 0] = t
    all_edges[:, :2, 1] = t[:, 1:]
    all_edges[:, 2, 1] = t[:, 0]
    all_edges = all_edges.reshape((-1, 2))
    all_edges.sort(axis = -1)  # first of pair now lower points index
    return np.unique(all_edges, axis = 0, return_counts = True)


def triangles_using_point(t, point_index):
    """Returns list-like 1D int array of indices of triangles using vertex identified by point_index."""

    assert t.ndim == 2 and t.shape[1] == 3 and isinstance(point_index, int)
    mask = np.any(t == point_index, axis = -1)
    return np.where(mask)[0]


def triangles_using_edge(t, p1, p2):
    """Returns list-like 1D int array of indices of triangles using edge identified by pair of point indices."""

    assert t.ndim == 2 and t.shape[1] == 3 and p1 != p2
    mask = np.logical_and(np.any(t == p1, axis = -1), np.any(t == p2, axis = -1))
    return np.where(mask)[0]


def triangles_using_edges(t, edges):
    """Returns int array of shape (len(edges), 2) with indices of upto 2 triangles using each edge (-1 for unused)."""

    assert t.ndim == 2 and t.shape[1] == 3 and edges.ndim == 2 and edges.shape[1] == 2
    ti = np.full((len(edges), 2), -1, dtype = int)
    for i in range(len(edges)):
        te = triangles_using_edge(t, edges[i, 0], edges[i, 1])
        c = len(te)
        assert 0 <= c <= 2
        if c:
            ti[i, :c] = te
    return ti


def rim_edges(all_edges, edge_counts):
    """Returns a subset of all edges where the edge count is 1."""

    assert all_edges.ndim == 2 and all_edges.shape[1] == 2
    assert edge_counts.ndim == 1 and edge_counts.size == len(all_edges)
    return all_edges[edge_counts == 1, :]


def internal_edges(all_edges, edge_counts):
    """Returns a subset of all edges where the edge count is 2."""

    assert all_edges.ndim == 2 and all_edges.shape[1] == 2
    assert edge_counts.ndim == 1 and edge_counts.size == len(all_edges)
    return all_edges[edge_counts == 2, :]


def rims(all_rim_edges):
    """Returns edge index and points index lists of distinct rims.

    arguments:
        all_rim_edges (numpy int array of shape (N, 2)): edge point indices; as returned by rim_edges()

    returns:
        (list of arrays of rim edge indices, list of arrays of corresponding points indices)
        where arrays are 1D numpy int arrays; those of first list hold indices into rows of all_rim_edges;
        those of second list hold the corresponding points indices, both ordered in sequence of rim
    """

    assert all_rim_edges.ndim == 2 and all_rim_edges.shape[1] == 2
    edge_count = len(all_rim_edges)
    edges_used_count = 0
    used_mask = np.zeros(edge_count, dtype = bool)
    rim_edges_list = []
    rim_points_list = []

    while edges_used_count < edge_count:

        index = _first_false(used_mask)
        assert index < edge_count
        start = all_rim_edges[index, 0]
        next = all_rim_edges[index, 1]
        used_mask[index] = True
        edges_used_count += 1
        rei = [index]
        rpi = [start]

        while next != start:
            rpi.append(next)
            index, next_next = _find_unused(all_rim_edges, used_mask, next)
            assert index < edge_count, 'loop not closed when following rim edges'
            used_mask[index] = True
            edges_used_count += 1
            rei.append(index)
            next = next_next

        rim_edges_list.append(np.array(rei, dtype = int))
        rim_points_list.append(np.array(rpi, dtype = int))

    return rim_edges_list, rim_points_list


@njit  # pragma: no cover
def _find_unused(ap: np.ndarray, used_mask: np.ndarray, v: int):  # type: ignore
    """Finds the first unused occurence of v in pair list ap, returning index and paired value."""
    for idx, val in np.ndenumerate(ap[:, 0]):
        if val == v and not used_mask[idx[0]]:
            return idx[0], ap[idx[0], 1]
    for idx, val in np.ndenumerate(ap[:, 1]):
        if val == v and not used_mask[idx[0]]:
            return idx[0], ap[idx[0], 0]
    return ap.shape[0], -1


@njit  # pragma: no cover
def _first_false(array: np.ndarray) -> Optional[int]:  # type: ignore
    """Returns the index of the first False (or zero) value in the 1D array."""
    for idx, val in np.ndenumerate(array):
        if not val:
            return idx[0]
    return array.size


@njit  # pragma: no cover
def _first_match(array: np.ndarray, v: int) -> Optional[int]:  # type: ignore
    """Returns the index of the first occurrence of value v in the 1D array."""
    for idx, val in np.ndenumerate(array):
        if val == v:
            return idx[0]
    return array.size



"""read_nexus_fault.py: functions for reading Nexus fault definition data from an ascii file."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import os
import re
import unicodedata
import numpy as np
import pandas as pd


def load_nexus_fault_mult_table_from_list(file_as_list):
    """Reads a Nexus (!) format list of file contents containing one or more MULT keywords and returns a dataframe with the MULT rows."""

    def is_number(s):
        try:
            float(s)
            return True
        except ValueError:
            pass

        try:
            unicodedata.numeric(s)
            return True
        except (TypeError, ValueError):
            pass

        return False

    dfs = []

    num_tables = 0
    ISTABLE = False
    ISRECORD = False
    name = face = None
    grid = 'ROOT'  ## nexus default grid

    face_dict = {'TX': 'I', 'TY': 'J', 'TZ': 'K', 'TI': 'I', 'TJ': 'J', 'TK': 'K'}

    chunks = []
    for line in file_as_list:
        if len(line.strip()):
            if (not line.strip()[0] == '!') & (not line.strip()[0] == 'C'):
                line = line.partition('!')[0]  # removing trailing comments
                # line = line.partition('C')[0]  # removing trailing comments
                tokens = line.split()
                if ISTABLE:
                    if is_number(tokens[0]):
                        ISRECORD = True

                    if ISRECORD and (not is_number(tokens[0])):
                        data = chunks[0:]
                        d_elems = np.array([np.array(data[i].split()) for i in range(len(data))])
                        # fill empty elements with zero
                        lens = np.array([len(i) for i in d_elems])
                        # Mask of valid places in each row
                        mask = np.arange(lens.max()) < lens[:, None]
                        # Setup output array and put elements from data into masked positions
                        outdata = np.zeros(mask.shape, dtype = d_elems.dtype)
                        outdata[mask] = np.concatenate(d_elems)
                        df = pd.DataFrame(outdata)
                        for column in df.columns:
                            df[column] = pd.to_numeric(df[column], errors = 'ignore')
                        df.columns = ['i1', 'i2', 'j1', 'j2', 'k1', 'k2', 'mult']
                        df['grid'] = grid
                        df['name'] = name
                        df['face'] = face
                        dfs.append(df)
                        num_tables += 1

                        ISTABLE = False
                        ISRECORD = False
                        chunks = []

                if ISTABLE:
                    if re.match("(.*)GRID(.*)", tokens[0]):
                        if len(tokens) > 0:
                            grid = tokens[1]
                    elif re.match("(.*)FNAME(.*)", tokens[0]):
                        if len(tokens) > 0:
                            name = tokens[1]
                    else:
                        if re.match(r"^MULT$", tokens[0]):
                            ISTABLE = False
                            ISRECORD = False
                            chunks = []
                        else:
                            chunks.append(line.strip())

                if re.match(r"^MULT$", tokens[0]):
                    if len(tokens) > 0:
                        face = face_dict[tokens[1]]
                    if 'MINUS' in tokens:
                        face += '-'  # indicates data apply to 'negative' faces of specified cells
                    grid = 'ROOT'  # nexus default
                    name = 'NONE'
                    ISTABLE = True

    else:
        if ISTABLE:
            if ISRECORD:
                data = chunks[0:]
                d_elems = np.array([np.array(data[i].split()) for i in range(len(data))])
                # fill empty elements with zero
                lens = np.array([len(i) for i in d_elems])
                # Mask of valid places in each row
                mask = np.arange(lens.max()) < lens[:, None]
                # Setup output array and put elements from data into masked positions
                outdata = np.zeros(mask.shape, dtype = d_elems.dtype)
                outdata[mask] = np.concatenate(d_elems)
                df = pd.DataFrame(outdata)
                for column in df.columns:
                    df[column] = pd.to_numeric(df[column], errors = 'ignore')
                df.columns = ['i1', 'i2', 'j1', 'j2', 'k1', 'k2', 'mult']
                df['grid'] = grid
                df['name'] = name
                df['face'] = face
                dfs.append(df)
                num_tables += 1

                ISTABLE = False
                ISRECORD = False
                chunks = []

    if not dfs:
        return pd.DataFrame()

    fault_df = pd.concat(dfs).reset_index(drop = True)

    convert_dict = {'i1': int, 'i2': int, 'j1': int, 'j2': int, 'k1': int, 'k2': int, 'mult': float}
    fault_df = fault_df.astype(convert_dict)

    return fault_df


def load_nexus_fault_mult_table(file_name):
    """Reads a Nexus (!) format file containing one or more MULT keywords and returns a dataframe with the MULT rows."""

    with open(file_name) as f:
        file_as_list = f.readlines()

    return load_nexus_fault_mult_table_from_list(file_as_list)



"""A submodule containing functions for extracting grid information"""

import logging

import numpy as np

import resqpy.olio.fine_coarse as fc
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.property as rprop
import resqpy.grid
import resqpy.grid._intervals_info as grr_ii
import resqpy.grid._defined_geometry as grr_dg

log = logging.getLogger(__name__)


def extract_grid_parent(grid):
    """Returns the uuid of the parent grid for the supplied grid"""
    if grid.extent_kji is None:
        grid.extract_extent_kji()
    if grid.parent_grid_uuid is not None:
        return grid.parent_grid_uuid
    grid.parent_window = None  # FineCoarse cell index mapping info with respect to parent
    grid.is_refinement = None
    pw_node = rqet.find_tag(grid.root, 'ParentWindow')
    if pw_node is None:
        return None
    # load a FineCoarse object as parent_window attribute and set parent_grid_uuid attribute
    grid.parent_grid_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(pw_node, ['ParentGrid', 'UUID']))
    assert grid.parent_grid_uuid is not None
    parent_grid_root = grid.model.root(uuid = grid.parent_grid_uuid)
    if parent_grid_root is None:
        log.warning('parent grid not present in model, unable to treat as local grid')
        return None
    # etxract parent grid extent directly from xml to avoid risk of circular references
    parent_grid_extent_kji = np.array((rqet.find_tag_int(
        parent_grid_root, 'Nk'), rqet.find_tag_int(parent_grid_root, 'Nj'), rqet.find_tag_int(parent_grid_root, 'Ni')),
                                      dtype = int)
    parent_initials = []
    intervals_count_list = []
    parent_count_list_list = []
    child_count_list_list = []
    child_weight_list_list = []
    refining_flag = None  # gets set True if local grid is a refinement, False if a coarsening
    parent_box = np.zeros((2, 3), dtype = int)
    for axis in range(3):
        refining_flag = __process_axis(axis, child_count_list_list, child_weight_list_list, grid, intervals_count_list,
                                       parent_box, parent_count_list_list, parent_grid_extent_kji, parent_initials,
                                       pw_node, refining_flag)
    cell_overlap_node = rqet.find_tag(pw_node, 'CellOverlap')
    if cell_overlap_node is not None:
        log.warning('ignoring cell overlap information in grid relationship')
    omit_node = rqet.find_tag(pw_node, 'OmitParentCells')
    if omit_node is not None:
        log.warning('unable to handle parent cell omissions in local grid definition – ignoring')
    # todo: handle omissions

    if refining_flag is None:
        log.warning('local grid has no refinement nor coarsening – treating as a refined grid')
        refining_flag = True
    grid.is_refinement = refining_flag

    if refining_flag:  # local grid is a refinement
        __extract_refined_parent(child_count_list_list, child_weight_list_list, grid, intervals_count_list, parent_box,
                                 parent_count_list_list)

    else:  # local grid is a coarsening
        __extract_coarsening_parent(child_count_list_list, child_weight_list_list, grid, intervals_count_list,
                                    parent_box, parent_count_list_list)

    grid.parent_window.assert_valid()

    return grid.parent_grid_uuid


def __extract_coarsening_parent(child_count_list_list, child_weight_list_list, grid, intervals_count_list, parent_box,
                                parent_count_list_list):
    grid.parent_window = fc.FineCoarse(parent_box[1] - parent_box[0] + 1, grid.extent_kji, within_fine_box = parent_box)
    for axis in range(3):
        if intervals_count_list[axis] == 1:
            grid.parent_window.set_constant_ratio(axis)
            constant_ratio = (parent_box[1, axis] - parent_box[0, axis] + 1) // grid.extent_kji[axis]
            ratio_vector = None
        else:
            constant_ratio = None
            ratio_vector = parent_count_list_list[axis] // child_count_list_list[axis]
            grid.parent_window.set_ratio_vector(axis, ratio_vector)
        if child_weight_list_list[axis] is None:
            grid.parent_window.set_equal_proportions(axis)
        else:
            proportions_list = []
            place = 0
            for coarse_slice in range(grid.extent_kji[axis]):
                if ratio_vector is None:
                    proportions_list.append(np.array(child_weight_list_list[axis][place:place + constant_ratio]))
                    place += constant_ratio
                else:
                    proportions_list.append(
                        np.array(child_weight_list_list[axis][place:place + ratio_vector[coarse_slice]]))
                    place += ratio_vector[coarse_slice]
            grid.parent_window.set_proportions_list_of_vectors(axis, proportions_list)


def __extract_refined_parent(child_count_list_list, child_weight_list_list, grid, intervals_count_list, parent_box,
                             parent_count_list_list):
    grid.parent_window = fc.FineCoarse(grid.extent_kji,
                                       parent_box[1] - parent_box[0] + 1,
                                       within_coarse_box = parent_box)
    for axis in range(3):
        if intervals_count_list[axis] == 1:
            grid.parent_window.set_constant_ratio(axis)
            constant_ratio = grid.extent_kji[axis] // (parent_box[1, axis] - parent_box[0, axis] + 1)
            ratio_vector = None
        else:
            constant_ratio = None
            ratio_vector = child_count_list_list[axis] // parent_count_list_list[axis]
            grid.parent_window.set_ratio_vector(axis, ratio_vector)
        if child_weight_list_list[axis] is None:
            grid.parent_window.set_equal_proportions(axis)
        else:
            proportions_list = []
            place = 0
            for coarse_slice in range(parent_box[1, axis] - parent_box[0, axis] + 1):
                if ratio_vector is None:
                    proportions_list.append(np.array(child_weight_list_list[axis][place:place + constant_ratio]))
                    place += constant_ratio
                else:
                    proportions_list.append(
                        np.array(child_weight_list_list[axis][place:place + ratio_vector[coarse_slice]]))
                    place += ratio_vector[coarse_slice]
            grid.parent_window.set_proportions_list_of_vectors(axis, proportions_list)


def __process_axis(axis, child_count_list_list, child_weight_list_list, grid, intervals_count_list, parent_box,
                   parent_count_list_list, parent_grid_extent_kji, parent_initials, pw_node, refining_flag):
    regrid_node = rqet.find_tag(pw_node, 'KJI'[axis] + 'Regrid')
    assert regrid_node is not None
    pii = rqet.find_tag_int(regrid_node, 'InitialIndexOnParentGrid')
    assert pii is not None and 0 <= pii < parent_grid_extent_kji[axis]
    parent_initials.append(pii)
    parent_box[0, axis] = pii
    intervals_node = rqet.find_tag(regrid_node, 'Intervals')
    if intervals_node is None:  # implicit one-to-one mapping
        intervals_count_list.append(1)
        parent_count_list_list.append(np.array(grid.extent_kji[axis], dtype = int))
        parent_box[1, axis] = parent_box[0, axis] + grid.extent_kji[axis] - 1
        assert parent_box[1, axis] < parent_grid_extent_kji[axis]
        child_count_list_list.append(np.array(grid.extent_kji[axis], dtype = int))
        child_weight_list_list.append(None)
    else:
        intervals_info = grr_ii.IntervalsInfo()
        intervals_count = rqet.find_tag_int(intervals_node, 'IntervalCount')
        assert intervals_count is not None and intervals_count > 0
        pcpi_node = rqet.find_tag(intervals_node, 'ParentCountPerInterval')
        assert pcpi_node is not None
        h5_key_pair = grid.model.h5_uuid_and_path_for_node(pcpi_node)
        assert h5_key_pair is not None
        grid.model.h5_array_element(h5_key_pair,
                                    index = None,
                                    cache_array = True,
                                    object = intervals_info,
                                    array_attribute = 'parent_count_per_interval',
                                    dtype = 'int')
        assert hasattr(intervals_info, 'parent_count_per_interval')
        assert intervals_info.parent_count_per_interval.ndim == 1 and intervals_info.parent_count_per_interval.size == intervals_count
        parent_box[1, axis] = parent_box[0, axis] + np.sum(intervals_info.parent_count_per_interval) - 1
        assert parent_box[1, axis] < parent_grid_extent_kji[axis]
        ccpi_node = rqet.find_tag(intervals_node, 'ChildCountPerInterval')
        assert ccpi_node is not None
        h5_key_pair = grid.model.h5_uuid_and_path_for_node(ccpi_node)
        assert h5_key_pair is not None
        grid.model.h5_array_element(h5_key_pair,
                                    index = None,
                                    cache_array = True,
                                    object = intervals_info,
                                    array_attribute = 'child_count_per_interval',
                                    dtype = 'int')
        assert hasattr(intervals_info, 'child_count_per_interval')
        assert intervals_info.child_count_per_interval.ndim == 1 and intervals_info.child_count_per_interval.size == intervals_count
        assert np.sum(intervals_info.child_count_per_interval) == grid.extent_kji[
            axis]  # assumes both local and parent grids are IjkGrids
        for interval in range(intervals_count):
            if intervals_info.child_count_per_interval[interval] == intervals_info.parent_count_per_interval[interval]:
                continue  # one-to-one
            if refining_flag is None:
                refining_flag = (intervals_info.child_count_per_interval[interval] >
                                 intervals_info.parent_count_per_interval[interval])
            assert refining_flag == (intervals_info.child_count_per_interval[interval] >
                                     intervals_info.parent_count_per_interval[interval]), \
                'mixture of refining and coarsening in one local grid – allowed by RESQML but not handled by this code'
            if refining_flag:
                assert intervals_info.child_count_per_interval[interval] % intervals_info.parent_count_per_interval[
                    interval] == 0, \
                    'within a single refinement interval, fine and coarse cell boundaries are not obviously aligned'
            else:
                assert intervals_info.parent_count_per_interval[interval] % intervals_info.child_count_per_interval[
                    interval] == 0, \
                    'within a single coarsening interval, fine and coarse cell boundaries are not obviously aligned'
        ccw_node = rqet.find_tag(intervals_node, 'ChildCellWeights')
        if ccw_node is None:
            intervals_info.child_cell_weights = None
        else:
            h5_key_pair = grid.model.h5_uuid_and_path_for_node(ccw_node)
            assert h5_key_pair is not None
            grid.model.h5_array_element(h5_key_pair,
                                        index = None,
                                        cache_array = True,
                                        object = intervals_info,
                                        array_attribute = 'child_cell_weights',
                                        dtype = 'float')
            assert hasattr(intervals_info, 'child_cell_weights')
            assert intervals_info.child_cell_weights.ndim == 1 and intervals_info.child_cell_weights.size == \
                   grid.extent_kji[
                       axis]
        intervals_count_list.append(intervals_count)
        parent_count_list_list.append(intervals_info.parent_count_per_interval)
        child_count_list_list.append(intervals_info.child_count_per_interval)
        child_weight_list_list.append(intervals_info.child_cell_weights)
    return refining_flag


def extract_extent_kji(grid):
    """Returns the grid extent; for IJK grids this is a 3 integer numpy array, order is Nk, Nj, Ni.

    returns:
       numpy int array of shape (3,) being number of cells in k, j & i axes respectively;
       the return value is cached in attribute extent_kji, which can alternatively be referenced
       directly by calling code as the value is set from xml on initialisation
    """

    if grid.extent_kji is not None:
        return grid.extent_kji
    grid.extent_kji = np.ones(3, dtype = 'int')  # todo: handle other varieties of grid
    grid.extent_kji[0] = int(rqet.find_tag(grid.root, 'Nk').text)
    grid.extent_kji[1] = int(rqet.find_tag(grid.root, 'Nj').text)
    grid.extent_kji[2] = int(rqet.find_tag(grid.root, 'Ni').text)
    return grid.extent_kji


def extract_grid_is_right_handed(grid):
    """Returns boolean indicating whether grid IJK axes are right handed, as stored in xml.

    returns:
       boolean: True if grid is right handed; False if left handed

    notes:
       this is the actual handedness of the IJK indexing of grid cells;
       the coordinate reference system has its own implicit handedness for xyz axes;
       Nexus requires the IJK space to be righthanded so if it is not, the handedness of the xyz space is
       falsified when exporting for Nexus (as Nexus allows xyz to be right or lefthanded and it is the
       handedness of the IJK space with respect to the xyz space that matters)
    """

    if grid.grid_is_right_handed is not None:
        return grid.grid_is_right_handed
    rh_node = grid.resolve_geometry_child('GridIsRighthanded')
    if rh_node is None:
        return None
    grid.grid_is_right_handed = (rh_node.text.lower() == 'true')
    return grid.grid_is_right_handed


def extract_k_direction_is_down(grid):
    """Returns boolean indicating whether increasing K indices are generally for deeper cells, as stored in xml.

    returns:
       boolean: True if increasing K generally indicates increasing depth

    notes:
       resqml allows layers to fold back over themselves, so the relationship between k and depth might not
       be monotonic;
       higher level code sometimes requires k to increase with depth;
       independently of this, z values may increase upwards or downwards in a coordinate reference system;
       this method does not modify the grid_is_righthanded indicator
    """

    if grid.k_direction_is_down is not None:
        return grid.k_direction_is_down
    k_dir_node = grid.resolve_geometry_child('KDirection')
    if k_dir_node is None:
        return None
    grid.k_direction_is_down = (k_dir_node.text.lower() == 'down')
    return grid.k_direction_is_down


def extract_geometry_time_index(grid):
    """Returns integer time index, or None, for the grid geometry, as stored in xml for dynamic geometries.

    notes:
       if the value is not None, it represents the time index as stored in the xml, or the time index as
       updated when setting the node points from a points property
    """
    if grid.time_index is not None and grid.time_series_uuid is not None:
        return grid.time_index
    grid.time_index = None
    grid.time_series_uuid = None
    ti_node = grid.resolve_geometry_child('TimeIndex')
    if ti_node is None:
        return None
    grid.time_index = rqet.find_tag_int(ti_node, 'Index')
    grid.time_series_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(ti_node, ['TimeSeries', 'UUID']))
    return grid.time_index


def extract_crs_uuid(grid):
    """Returns uuid for coordinate reference system, as stored in geometry xml tree.

    returns:
       uuid.UUID object
    """

    if grid.crs_uuid is not None:
        return grid.crs_uuid
    crs_root = grid.resolve_geometry_child('LocalCrs')
    uuid_str = rqet.find_tag_text(crs_root, 'UUID')
    if not uuid_str and hasattr(grid, 'extra_metadata'):
        uuid_str = grid.extra_metadata.get('crs uuid')
    if uuid_str:
        grid.crs_uuid = bu.uuid_from_string(uuid_str)
    return grid.crs_uuid


def extract_crs_root(grid):
    """Returns root in parent model xml parts forest of coordinate reference system used by this grid geomwtry.

    returns:
       root node in xml tree for coordinate reference system

    note:
       resqml allows a part to refer to another part that is not actually present in the same epc package;
       in practice, the crs is a tiny part and has always been included in datasets encountered so far;
       if the crs is not present, this method will return None (I think)
    """

    if grid.crs_root is not None:
        return grid.crs_root
    crs_uuid = grid.extract_crs_uuid()
    if crs_uuid is None:
        return None
    grid.crs_root = grid.model.root(uuid = crs_uuid)
    return grid.crs_root


def extract_pillar_shape(grid):
    """Returns string indicating whether whether pillars are curved, straight, or vertical as stored in xml.

    returns:
       string: either 'curved', 'straight' or 'vertical'

    note:
       resqml datasets often have 'curved', even when the pillars are actually 'vertical' or 'straight';
       use actual_pillar_shape() method to determine the shape from the actual xyz points data
    """

    if grid.pillar_shape is not None:
        return grid.pillar_shape
    ps_node = grid.resolve_geometry_child('PillarShape')
    if ps_node is None:
        return None
    grid.pillar_shape = ps_node.text
    return grid.pillar_shape


def extract_has_split_coordinate_lines(grid):
    """Returns boolean indicating whether grid geometry has any split coordinate lines (split pillars, ie. faults).

    returns:
       boolean: True if the grid has one or more split pillars; False if all pillars are unsplit

    notes:
       the return value is based on the array elements present in the xml tree, unless it has already been
       determined;
       resqml ijk grids with split coordinate lines have extra arrays compared to unfaulted grids, and the main
       Points array is indexed differently: [k', pillar_index, xyz] instead of [k', j', i', xyz] (where k', j', i'
       range of nk+k_gaps+1, nj+1, ni+1 respectively)
    """

    if grid.has_split_coordinate_lines is not None:
        return grid.has_split_coordinate_lines
    split_node = grid.resolve_geometry_child('SplitCoordinateLines')
    grid.has_split_coordinate_lines = (split_node is not None)
    if split_node is not None:
        grid.split_pillars_count = int(rqet.find_tag(split_node, 'Count').text.strip())
    return grid.has_split_coordinate_lines


def extract_k_gaps(grid):
    """Returns information about gaps (voids) between layers in the grid.

    returns:
       (int, numpy bool array, numpy int array) being the number of gaps between layers;
       a 1D bool array of extent nk-1 set True where there is a gap below the layer; and
       a 1D int array being the k index to actually use in the points data for each layer k0

    notes:
       all returned elements are stored as attributes in the grid object; int and bool array elements
       will be None if there are no k gaps; each k gap implies an extra element in the points data for
       each pillar; when wanting to index k interfaces (horizons) rather than layers, the last of the
       returned values can be used to index the k axis of the points data to yield the top face of the
       layer and the successor in k will always index the basal face of the same layer
    """

    if grid.k_gaps is not None:
        return grid.k_gaps, grid.k_gap_after_array, grid.k_raw_index_array
    grid.k_gaps = rqet.find_nested_tags_int(grid.root, ['KGaps', 'Count'])
    if grid.k_gaps:
        k_gap_after_root = rqet.find_nested_tags(grid.root, ['KGaps', 'GapAfterLayer'])
        assert k_gap_after_root is not None
        bool_array_type = rqet.node_type(k_gap_after_root)
        assert bool_array_type == 'BooleanHdf5Array'  # could be a constant array but not handled by this code
        h5_key_pair = grid.model.h5_uuid_and_path_for_node(k_gap_after_root)
        assert h5_key_pair is not None
        grid.model.h5_array_element(h5_key_pair,
                                    index = None,
                                    cache_array = True,
                                    object = grid,
                                    array_attribute = 'k_gap_after_array',
                                    dtype = 'bool')
        assert hasattr(grid, 'k_gap_after_array')
        assert grid.k_gap_after_array.ndim == 1 and grid.k_gap_after_array.size == grid.nk - 1
        grid._set_k_raw_index_array()
    else:
        grid.k_gap_after_array = None
        grid.k_raw_index_array = np.arange(grid.nk, dtype = int)
    return grid.k_gaps, grid.k_gap_after_array, grid.k_raw_index_array


def extract_stratigraphy(grid):
    """Loads stratigraphic information from xml."""

    grid.stratigraphic_column_rank_uuid = None
    grid.stratigraphic_units = None
    strata_node = rqet.find_tag(grid.root, 'IntervalStratigraphicUnits')
    if strata_node is None:
        return
    grid.stratigraphic_column_rank_uuid = \
        bu.uuid_from_string(rqet.find_nested_tags_text(strata_node, ['StratigraphicOrganization', 'UUID']))
    assert grid.stratigraphic_column_rank_uuid is not None
    unit_indices_node = rqet.find_tag(strata_node, 'UnitIndices')
    h5_key_pair = grid.model.h5_uuid_and_path_for_node(unit_indices_node)
    grid.model.h5_array_element(h5_key_pair,
                                index = None,
                                cache_array = True,
                                object = grid,
                                array_attribute = 'stratigraphic_units',
                                dtype = 'int')
    assert len(grid.stratigraphic_units) == grid.nk_plus_k_gaps


def extract_children(grid):
    """Looks for LGRs related to this grid and sets the local_grid_uuid_list attribute."""
    assert grid.uuid is not None
    if grid.local_grid_uuid_list is not None:
        return grid.local_grid_uuid_list
    grid.local_grid_uuid_list = []
    related_grid_roots = grid.model.roots(obj_type = 'IjkGridRepresentation', related_uuid = grid.uuid)
    if related_grid_roots is not None:
        for related_root in related_grid_roots:
            parent_uuid = rqet.find_nested_tags_text(related_root, ['ParentWindow', 'ParentGrid', 'UUID'])
            if parent_uuid is None:
                continue
            parent_uuid = bu.uuid_from_string(parent_uuid)
            if bu.matching_uuids(grid.uuid, parent_uuid):
                grid.local_grid_uuid_list.append(parent_uuid)
    return grid.local_grid_uuid_list


def extract_property_collection(grid):
    """Load grid property collection object holding lists of all properties in model that relate to this grid.

    returns:
       resqml_property.GridPropertyCollection object

    note:
       a reference to the grid property collection is cached in this grid object; if the properties change,
       for example by generating some new properties, the property_collection attribute of the grid object
       would need to be reset to None elsewhere before calling this method again
    """

    if grid.property_collection is not None:
        return grid.property_collection
    grid.property_collection = rprop.GridPropertyCollection(grid = grid)
    return grid.property_collection


def extract_inactive_mask(grid, check_pinchout = False):
    """Returns boolean numpy array indicating which cells are inactive, if (in)active property found in this grid.

    returns:
       numpy array of booleans, of shape (nk, nj, ni) being True for cells which are inactive; False for active

    note:
       RESQML does not have a built-in concept of inactive (dead) cells, though the usage guide advises to use a
       discrete property with a local property kind of 'active'; this resqpy code can maintain an 'inactive'
       attribute for the grid object, which is a boolean numpy array indicating which cells are inactive
    """

    if grid.inactive is not None and not check_pinchout:
        return grid.inactive
    geom_defined = grr_dg.cell_geometry_is_defined_ref(grid)
    if grid.inactive is None:
        if geom_defined is None or geom_defined is True:
            grid.inactive = np.zeros(tuple(grid.extent_kji))  # ie. all active
        else:
            grid.inactive = np.logical_not(grr_dg.cell_geometry_is_defined_ref(grid))
    if check_pinchout:
        grid.inactive = np.logical_or(grid.inactive, grid.pinched_out())
    gpc = grid.extract_property_collection()
    if gpc is None:
        grid.all_inactive = np.all(grid.inactive)
        return grid.inactive
    active_gpc = rprop.GridPropertyCollection()
    # note: use of bespoke (local) property kind 'active' as suggested in resqml usage guide
    active_gpc.inherit_parts_selectively_from_other_collection(other = gpc,
                                                               property_kind = 'active',
                                                               indexable = 'cells',
                                                               continuous = False)
    active_parts = active_gpc.parts()
    if len(active_parts) > 1:
        # try further filtering based on grid's time index data (or filtering out time based arrays)
        grid.extract_geometry_time_index()
        if grid.time_index is not None and grid.time_series_uuid is not None:
            active_gpc = rprop.selective_version_of_collection(active_gpc,
                                                               time_index = grid.time_index,
                                                               time_series_uuid = grid.time_series_uuid)
        else:
            active_parts = []
            for part in active_gpc.parts():
                if active_gpc.time_series_uuid_for_part(part) is None and active_gpc.time_index_for_part(part) is None:
                    active_parts.append(part)
    if len(active_parts) > 0:
        if len(active_parts) > 1:
            log.warning('more than one property found with bespoke kind "active", using last encountered')
        active_part = active_parts[-1]
        active_array = active_gpc.cached_part_array_ref(active_part, dtype = 'bool')
        grid.inactive = np.logical_or(grid.inactive, np.logical_not(active_array))
        grid.active_property_uuid = active_gpc.uuid_for_part(active_part)
        active_gpc.uncache_part_array(active_part)
    else:  # for backward compatibility with earlier versions of resqpy
        inactive_gpc = rprop.GridPropertyCollection()
        inactive_gpc.inherit_parts_selectively_from_other_collection(other = gpc,
                                                                     property_kind = 'code',
                                                                     facet_type = 'what',
                                                                     facet = 'inactive')
        if inactive_gpc.number_of_parts() == 1:
            inactive_part = inactive_gpc.parts()[0]
            inactive_array = inactive_gpc.cached_part_array_ref(inactive_part, dtype = 'bool')
            grid.inactive = np.logical_or(grid.inactive, inactive_array)
            inactive_gpc.uncache_part_array(inactive_part)

    grid.all_inactive = np.all(grid.inactive)
    return grid.inactive


def extent_kji_from_root(root_node):
    """Returns kji extent as stored in xml."""

    return (rqet.find_tag_int(root_node, 'Nk'), rqet.find_tag_int(root_node, 'Nj'), rqet.find_tag_int(root_node, 'Ni'))


def set_k_direction_from_points(grid):
    """Sets the K direction indicator based on z direction and mean z values for top and base.

    note:
       this method does not modify the grid_is_righthanded indicator
    """

    p = grid.points_ref(masked = False)
    grid.k_direction_is_down = True  # arbitrary default
    if p is not None:
        diff = np.nanmean(p[-1] - p[0])
        if not np.isnan(diff):
            grid.k_direction_is_down = ((diff >= 0.0) == grid.z_inc_down())
    return grid.k_direction_is_down



"""Functions relating to intersection of resqml grid with surface or trajectory objects."""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.intersection as meet
import resqpy.olio.vector_utilities as vec
import resqpy.surface as rqs


def generate_untorn_surface_for_layer_interface(grid,
                                                k0 = 0,
                                                ref_k_faces = 'top',
                                                quad_triangles = True,
                                                border = None):
    """Returns a Surface object generated from the grid layer interface points after any faults are 'healed'.

    arguments:
       grid (grid.Grid object): the grid object from which a layer interface is to be converted to a surface
       k0 (int): the layer number (zero based) to be used
       ref_k_faces (string): either 'top' (the default) or 'base', indicating whether the top or the base
          interface of the layer is to be used
       quad_triangles (boolean, optional, default True): if True, 4 triangles are used to represent each cell k face,
          which gives a unique solution with a shared node of the 4 triangles at the mean point of the 4 corners of
          the face; if False, only 2 triangles are used, which gives a non-unique solution
       border (float, optional): If given, an extra border row of quadrangles is added around the grid mesh

    returns:
       a resqml_surface.Surface object with a single triangulated patch

    notes:
       The resulting surface is assigned to the same model as grid, though xml is not generated and hdf5 is not
       written.
       If a border is specified and the outer grid cells have non-parallel edges, the resulting mesh might be
       messed up.
    """

    surf = rqs.Surface(grid.model)
    kp = 1 if ref_k_faces == 'base' else 0
    mesh = grid.horizon_points(ref_k0 = k0, heal_faults = True, kp = kp)
    if border is None or border <= 0.0:
        surf.set_from_irregular_mesh(mesh, quad_triangles = quad_triangles)
    else:
        #      origin = np.mean(mesh, axis = (0, 1))
        skirted_mesh = np.empty((mesh.shape[0] + 2, mesh.shape[1] + 2, 3))
        skirted_mesh[1:-1, 1:-1, :] = mesh
        # fill border values (other than corners)
        # yaml: disable
        for j in range(1, mesh.shape[0] + 1):
            skirted_mesh[j, 0, :] =  \
                skirted_mesh[j, 1] + border * vec.unit_vector(skirted_mesh[j, 1] - skirted_mesh[j, 2])
            skirted_mesh[j, -1, :] =  \
                skirted_mesh[j, -2] + border * vec.unit_vector(skirted_mesh[j, -2] - skirted_mesh[j, -3])
        for i in range(1, mesh.shape[1] + 1):
            skirted_mesh[0, i, :] =  \
                skirted_mesh[1, i] + border * vec.unit_vector(skirted_mesh[1, i] - skirted_mesh[2, i])
            skirted_mesh[-1, i, :] =  \
                skirted_mesh[-2, i] + border * vec.unit_vector(skirted_mesh[-2, i] - skirted_mesh[-3, i])
        # yaml: enable
        # fill in corner values
        skirted_mesh[0, 0, :] = skirted_mesh[0, 1] + skirted_mesh[1, 0] - skirted_mesh[1, 1]
        skirted_mesh[0, -1, :] = skirted_mesh[0, -2] + skirted_mesh[1, -1] - skirted_mesh[1, -2]
        skirted_mesh[-1, 0, :] = skirted_mesh[-1, 1] + skirted_mesh[-2, 0] - skirted_mesh[-2, 1]
        skirted_mesh[-1, -1, :] = skirted_mesh[-1, -2] + skirted_mesh[-2, -1] - skirted_mesh[-2, -2]
        surf.set_from_irregular_mesh(skirted_mesh, quad_triangles = quad_triangles)

    return surf


def generate_torn_surface_for_layer_interface(grid, k0 = 0, ref_k_faces = 'top', quad_triangles = True):
    """Returns a Surface object generated from the grid layer interface points.

    arguments:
       grid (grid.Grid object): the grid object from which a layer interface is to be converted to a surface
       k0 (int): the layer number (zero based) to be used
       ref_k_faces (string): either 'top' (the default) or 'base', indicating whether the top or the base
          interface of the layer is to be used
       quad_triangles (boolean, optional, default True): if True, 4 triangles are used to represent each cell k face,
          which gives a unique solution with a shared node of the 4 triangles at the mean point of the 4 corners of
          the face; if False, only 2 triangles are used, which gives a non-unique solution

    returns:
       a resqml_surface.Surface object with a single triangulated patch

    notes:
       The resulting surface is assigned to the same model as grid, though xml is not generated and hdf5 is not
       written.
       Strictly, the RESQML business rules for a triangulated surface require a separate patch for areas of the
       surface which are not joined; therefore, if fault tears cut off one area of the surface (eg. a fault running
       fully across the grid), then more than one patch should be generated; however, at present the code uses a
       single patch regardless.
    """

    surf = rqs.Surface(grid.model)
    kp = 1 if ref_k_faces == 'base' else 0
    mesh = grid.split_horizon_points(ref_k0 = k0, kp = kp)
    surf.set_from_torn_mesh(mesh, quad_triangles = quad_triangles)

    return surf


def generate_torn_surface_for_x_section(grid,
                                        axis,
                                        ref_slice0 = 0,
                                        plus_face = False,
                                        quad_triangles = True,
                                        as_single_layer = False):
    """Returns a Surface object generated from the grid cross section points.

    arguments:
       grid (grid.Grid object): the grid object from which a cross section is to be converted to a surface
       axis (string): 'I' or 'J' being the axis of the cross-sectional slice (ie. dimension being dropped)
       ref_slice0 (int, default 0): the reference value for indices in I or J (as defined in axis)
       plus_face (boolean, default False): if False, negative face is used; if True, positive
       quad_triangles (boolean, default True): if True, 4 triangles are used to represent each cell face, which
          gives a unique solution with a shared node of the 4 triangles at the mean point of the 4 corners of
          the face; if False, only 2 triangles are used, which gives a non-unique solution
       as_single_layer (boolean, default False): if True, the top points from the top layer are used together
          with the basal points from the base layer, to effect a single layer equivalent cross section surface

    returns:
       a resqml_surface.Surface object with a single triangulated patch

    notes:
       The resulting surface is assigned to the same model as grid, though xml is not generated and hdf5 is not
       written.
       Strictly, the RESQML business rules for a triangulated surface require a separate patch for areas of the
       surface which are not joined; therefore, a fault running down through the grid should result in separate
       patches; however, at present the code uses a single patch regardless.
    """

    assert axis.upper() in ['I', 'J']

    if grid.k_gaps is None or grid.k_gaps == 0:
        x_sect_points = grid.split_x_section_points(axis, ref_slice0 = ref_slice0, plus_face = plus_face)
        if as_single_layer:
            shape = np.array(x_sect_points.shape)
            shape[0] = 1
            x_sect_top = x_sect_points[0].reshape(tuple(shape))
            x_sect_base = x_sect_points[-1].reshape(tuple(shape))
        else:
            x_sect_top = x_sect_points[:-1]
            x_sect_base = x_sect_points[1:]
        x_sect_mesh = np.stack((x_sect_top, x_sect_base), axis = 2)
    else:
        x_sect_mesh = grid.split_gap_x_section_points(axis, ref_slice0 = ref_slice0, plus_face = plus_face)

    surf = rqs.Surface(grid.model)
    surf.set_from_torn_mesh(x_sect_mesh, quad_triangles = quad_triangles)

    return surf


def generate_untorn_surface_for_x_section(grid,
                                          axis,
                                          ref_slice0 = 0,
                                          plus_face = False,
                                          quad_triangles = True,
                                          as_single_layer = False):
    """Returns a Surface object generated from the grid cross section points for an unfaulted grid.

    arguments:
       grid (grid.Grid object): the grid object from which a cross section is to be converted to a surface
       axis (string): 'I' or 'J' being the axis of the cross-sectional slice (ie. dimension being dropped)
       ref_slice0 (int, default 0): the reference value for indices in I or J (as defined in axis)
       plus_face (boolean, default False): if False, negative face is used; if True, positive
       quad_triangles (boolean, default True): if True, 4 triangles are used to represent each cell face, which
          gives a unique solution with a shared node of the 4 triangles at the mean point of the 4 corners of
          the face; if False, only 2 triangles are used, which gives a non-unique solution
       as_single_layer (boolean, default False): if True, the top points from the top layer are used together
          with the basal points from the base layer, to effect a single layer equivalent cross section surface

    returns:
       a resqml_surface.Surface object with a single triangulated patch

    notes:
       The resulting surface is assigned to the same model as grid, though xml is not generated and hdf5 is not
       written.
       Strictly, the RESQML business rules for a triangulated surface require a separate patch for areas of the
       surface which are not joined; therefore, a fault running down through the grid should result in separate
       patches; however, at present the code uses a single patch regardless.
    """

    assert axis.upper() in ['I', 'J']

    x_sect_points = grid.unsplit_x_section_points(axis, ref_slice0 = ref_slice0, plus_face = plus_face)
    if as_single_layer:
        shape = np.array(x_sect_points.shape)
        shape[0] = 1
        x_sect_top = x_sect_points[0]
        x_sect_base = x_sect_points[-1]
        x_sect_mesh = np.stack((x_sect_top, x_sect_base), axis = 0)
    else:
        x_sect_mesh = x_sect_points

    log.debug(f'x_sect_mesh.shape: {x_sect_mesh.shape}; grid.extent_kji: {grid.extent_kji}')

    surf = rqs.Surface(grid.model)
    surf.set_from_irregular_mesh(x_sect_mesh, quad_triangles = quad_triangles)

    return surf


def point_is_within_cell(xyz, grid, kji0, cell_surface = None, false_on_pinchout = True):
    """Returns True if point xyz is within cell kji0, but not on its surface."""

    if false_on_pinchout and grid.pinched_out(kji0, cache_pinchout_array = False):
        return False
    if cell_surface is None:
        cp = grid.corner_points(kji0)
        cell_surface = rqs.Surface(grid.model)
        cell_surface.set_to_single_cell_faces_from_corner_points(cp, quad_triangles = True)
    t, p = cell_surface.triangles_and_points()
    triangles = p[t]
    centre = grid.centre_point(kji0)
    line_v = centre - xyz
    intersects = meet.line_triangles_intersects(xyz, line_v, triangles, line_segment = True)
    return np.all(np.isnan(intersects))


def create_column_face_mesh_and_surface(grid, col_ji0, axis, polarity, quad_triangles = True, as_single_layer = False):
    """Creates a Mesh and corresponding Surface representing a column face.

    arguments:
       grid (grid.Grid object)
       col_ji0 (int pair): the column indices, zero based
       axis (int): 1 for J face, 2 fo I face
       polarity (int): 0 for negative face, 1 for positive
       quad_triangles (boolean, default True): if True, 4 triangles are used per cell face; if False, 2 triangles
       as_single_layer (boolean, default False): if True, only the top and basal points are used, with the results being
          equivalent to the grid being treated as a single layer

    returns:
       surface.Mesh, surface.Surface (or None, surface.Surface if grid has k gaps)
    """

    assert axis in (1, 2)

    col_pm = grid.create_column_pillar_mapping()[col_ji0[0], col_ji0[1]]
    if axis == 1:  # J face
        pillar_index_pair = col_pm[polarity, :]
    else:  # I face
        pillar_index_pair = col_pm[:, polarity]
    if grid.k_gaps:
        points = grid.points_ref(masked = False).reshape(grid.nk_plus_k_gaps + 1, -1, 3)
    else:
        points = grid.points_ref(masked = False).reshape(grid.nk + 1, -1, 3)
    # note, here col_face_xyz is indexed by (j or i, k, xyz) whereas elsewhere (k, j or i, xyz) would be more typical
    # this protocol needs to align with re-use of Surface.column_for_triangle_index() method for layer identification

    if not as_single_layer and grid.k_gaps:
        col_face_mesh = None
        col_face_surface = rqs.Surface(grid.model)
        mesh = np.empty((1, grid.nk, 2, 2, 3))
        mesh[0, :, 0, 0, :] = points[grid.k_raw_index_array, pillar_index_pair[0], :]
        mesh[0, :, 1, 0, :] = points[grid.k_raw_index_array, pillar_index_pair[1], :]
        mesh[0, :, 0, 1, :] = points[grid.k_raw_index_array + 1, pillar_index_pair[0], :]
        mesh[0, :, 1, 1, :] = points[grid.k_raw_index_array + 1, pillar_index_pair[1], :]
        col_face_surface.set_from_torn_mesh(mesh, quad_triangles = quad_triangles)

    else:
        if as_single_layer:
            col_face_xyz = np.empty((2, 2, 3))
            col_face_xyz[0, 0] = points[0, pillar_index_pair[0]]
            col_face_xyz[0, 1] = points[-1, pillar_index_pair[0]]
            col_face_xyz[1, 0] = points[0, pillar_index_pair[1]]
            col_face_xyz[1, 1] = points[-1, pillar_index_pair[1]]
        else:
            col_face_xyz = np.empty((2, grid.nk + 1, 3))
            col_face_xyz[0] = points[:, pillar_index_pair[0]]
            col_face_xyz[1] = points[:, pillar_index_pair[1]]
        col_face_mesh = rqs.Mesh(grid.model, xyz_values = col_face_xyz, crs_uuid = grid.crs_uuid)
        title = 'column face for j0,i0: ' + str(col_ji0[0]) + ',' + str(
            col_ji0[1]) + ' face ' + 'KJI'[axis] + '-+'[polarity]
        col_face_surface = rqs.Surface(grid.model, mesh = col_face_mesh, quad_triangles = quad_triangles, title = title)

    return col_face_mesh, col_face_surface



"""_add_ab_properties.py: Module to add binary grid properties to an existing RESQMl grid object."""

import logging

log = logging.getLogger(__name__)

import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.property as rp


def add_ab_properties(epc_file, grid_uuid = None, ext_uuid = None, ab_property_list = None):
    """Import a list of pure binary property array files as grid properties.

    arguments:
        epc_file (str): path of existing resqml epc to be added to
        grid_uuid (UUID, optional): the uuid of the grid to receive the properties; required if more than one grid present
        ext_uuid (UUID, optional): the uuid of the hdf5 extension part to use for the arrays; recommended to leave as None
        ab_property_list (list of tuples): each entry contains:
            (file_name, keyword, property_kind, facet_type, facet, uom, time_index, null_value, discrete, realization)

    returns:
        Model, with the new properties added, with hdf5 and epc fully updated
    """

    assert ab_property_list, 'property list is empty or missing'

    model = rq.Model(epc_file = epc_file)
    if grid_uuid is None:
        grid_node = model.root_for_ijk_grid()  # will raise an exception if Model has more than 1 grid
        assert grid_node is not None, 'grid not found in model'
        grid_uuid = rqet.uuid_for_part_root(grid_node)
    grid = grr.any_grid(parent_model = model, uuid = grid_uuid, find_properties = False)

    if ext_uuid is None:
        ext_node = rqet.find_nested_tags(grid.geometry_root, ['Points', 'Coordinates', 'HdfProxy', 'UUID'])
        if ext_node is not None:
            ext_uuid = bu.uuid_from_string(ext_node.text.strip())

    #  ab_property_list: list of (filename, keyword, property_kind, facet_type, facet, uom, time_index, null_value, discrete, realization)
    prop_import_collection = rp.GridPropertyCollection()
    prop_import_collection.set_grid(grid)
    for (p_filename, p_keyword, p_property_kind, p_facet_type, p_facet, p_uom, p_time_index, p_null_value, p_discrete,
         p_realization) in ab_property_list:
        prop_import_collection.import_ab_property_to_cache(p_filename,
                                                           p_keyword,
                                                           grid.extent_kji,
                                                           discrete = p_discrete,
                                                           uom = p_uom,
                                                           time_index = p_time_index,
                                                           null_value = p_null_value,
                                                           property_kind = p_property_kind,
                                                           facet_type = p_facet_type,
                                                           facet = p_facet,
                                                           realization = p_realization)
        # todo: property_kind, facet_type & facet are not currently getting passed through the imported_list tuple in resqml_property

    if prop_import_collection is None:
        log.warning('no pure binary grid properties to import')
    else:
        log.info('number of pure binary grid property arrays: ' + str(prop_import_collection.number_of_imports()))

    # append to hdf5 file using arrays cached in grid property collection above
    hdf5_file = model.h5_file_name()
    log.debug('appending to hdf5 file: ' + hdf5_file)
    grid.write_hdf5_from_caches(hdf5_file,
                                mode = 'a',
                                geometry = False,
                                imported_properties = prop_import_collection,
                                write_active = False)
    # remove cached static property arrays from memory
    if prop_import_collection is not None:
        prop_import_collection.remove_all_cached_arrays()

    # add imported properties parts to model, building property parts list
    if prop_import_collection is not None and prop_import_collection.imported_list is not None:
        prop_import_collection.create_xml_for_imported_list_and_add_parts_to_model(ext_uuid)

    # mark model as modified
    model.set_modified()

    # store new version of model
    log.info('storing model with additional properties in epc file: ' + epc_file)
    model.store_epc(epc_file)

    return model



"""Mesh class based on RESQML Grid2dRepresentation class."""

# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company
# GOCAD is also a trademark of Emerson

import logging

log = logging.getLogger(__name__)

import warnings
import math as maths
import numpy as np

import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp
import resqpy.weights_and_measures as wam
import resqpy.surface
import resqpy.surface._base_surface as rqsb
import resqpy.surface._surface as rqss
from resqpy.olio.xml_namespaces import curly_namespace as ns
from resqpy.olio.zmap_reader import read_mesh


class Mesh(rqsb.BaseSurface):
    """Class covering meshes (lattices: surfaces where points form a 2D grid; RESQML obj_Grid2dRepresentation)."""

    resqml_type = 'Grid2dRepresentation'

    def __init__(self,
                 parent_model,
                 uuid = None,
                 mesh_file = None,
                 mesh_format = None,
                 mesh_flavour = 'explicit',
                 xyz_values = None,
                 nj = None,
                 ni = None,
                 origin = None,
                 dxyz_dij = None,
                 z_values = None,
                 z_supporting_mesh_uuid = None,
                 surface_role = 'map',
                 crs_uuid = None,
                 title = None,
                 originator = None,
                 extra_metadata = None):
        """Initialises a Mesh object from xml, or a regular mesh from arguments.

        arguments:
           parent_model (model.Model object): the model to which this Mesh object will be associated
           uuid (uuid.UUID, optional): the uuid of an existing RESQML obj_Grid2dRepresentation object from which
              this resqpy Mesh object is populated
           mesh_file (string, optional): file name, required if initialising from an RMS text or zmap+ ascii file
           mesh_format (string, optional): 'rms' or 'zmap', required if initialising from an ascii file
           mesh_flavour (string, default 'explicit'): required flavour when reading from a mesh file; one of:
              'explicit', 'regular' (z values discarded), 'reg&z', 'ref&z'
           xyz_values (numpy int array of shape (nj, ni, 3), optional): can be used to create an explicit
              mesh directly from the full array of points
           nj (int, optional): when generating a regular or 'ref&z' mesh, the number of nodes (NB. not 'cells')
              in the j axis of the regular mesh
           ni (int, optional): the number of nodes in the i axis of the regular or ref&z mesh
           origin (triple float, optional): the xyz origin of the regular mesh; use z value of zero if irrelevant
           dxyz_dij (numpy float array of shape (2, 3), optional): the xyz increment for each step in i and j axes;
              use z increments of zero if not applicable; eg. [[50.0, 0.0, 0.0], [0.0, 50.0, 0.0]] for mesh with
              50 (m or ft) spacing where the I axis aligns with x axis and the J axis aligns with y axis; first of
              the two triplets relates to the I axis
           z_values (numpy int array of shape (nj, ni), optional): z values used when creating a ref&z flavour
              mesh; z_supporting_mesh_uuid must also be supplied
           z_supporting_mesh_uuid (uuid.UUID, optional): used to specify the supporting mesh when creating a
              ref&z or reg&z flavour mesh; z_values must also be supplied
           surface_role (string, default 'map'): 'map' or 'pick'; ignored if uuid is not None
           crs_uuid (uuid.Uuid or string, optional): required if generating a regular mesh, the uuid of the crs
           title (str, optional): the citation title to use for a new mesh;
              ignored if uuid is not None
           originator (str, optional): the name of the person creating the mesh, defaults to login id;
              ignored if uuid is not None
           extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the mesh;
              ignored if uuid is not None

        returns:
           the newly created Mesh object

        notes:
           a mesh is a set of x,y,z (or x,y) points organised into a 2D lattice indexed by j,i; the z values are
           sometimes not applicable and then can be set to zero; 3 flavours of mesh are supported (the RESQML
           standard might allow for others): regular, where a constant xyz delta is applied with each step in i
           and j, starting from an origin, to yield a planar surface; explicit, where the full xyz (or xy) data
           is held as an array; 'ref & z', where another mesh is referred to for xy data, and z values are held
           in an array;
           there are 5 ways to initialise a Mesh object, in order of precedence:
           1. pass uuid to initialise from xml
           2. pass mesh_file, mesh_format and crs_uuid to load an explicit mesh from an ascii file
           3. pass xyz_values and crs_uuid to create an explicit mesh from a numpy array
           4. pass nj, ni, origin, dxyz_dij and crs_uuid to initialise a regular mesh directly
           5. pass z_values, z_supporting_mesh_uuid and crs_uuid to initialise a 'ref & z' mesh
           6. leave all optional arguments as None for an empty Mesh object

        :meta common:
        """

        assert surface_role in ['map', 'pick']

        self.surface_role = surface_role
        self.ni = None  # NB: these are the number of nodes (points) in the mesh, unlike 3D grids
        self.nj = None
        self.flavour = None  # 'explicit', 'regular', 'ref&z' or 'reg&z'
        self.full_array = None  # loaded on demand, shape (NJ, NI, 3 or 2) being xyz or xy points at each node
        self.explicit_h5_key_pair = None
        self.regular_origin = None  # xyz of origin of regular mesh
        self.regular_dxyz_dij = None  # numpy array of shape (2, 3) being xyz increment with each step in i, j
        self.ref_uuid = None
        self.ref_mesh = None
        self.ref_z_h5_key_pair = None
        # note: in this class, z values for ref&z meshes are held in the full_array (xy data will be duplicated in memory)
        self.crs_uuid = crs_uuid
        self.represented_interpretation_root = None

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is not None:
            pass

        elif mesh_file and mesh_format and crs_uuid is not None:
            self.__load_from_mesh_file(mesh_file, mesh_flavour, mesh_format, crs_uuid, ni, nj)

        elif xyz_values is not None and crs_uuid is not None:
            self.__load_from_xyz_values(xyz_values)

        else:
            self.__load_from_arguments(ni, nj, origin, dxyz_dij, z_values, z_supporting_mesh_uuid, crs_uuid)

        assert self.crs_uuid is not None
        if not self.title:
            self.title = 'mesh'

    @classmethod
    def from_regular_grid_column_property(cls, parent_model, grid_uuid, property_uuid):
        """Creates a reg&z flavoured Mesh from an aligned regular grid column property.

        arguments:
            parent_model (Model): the model that the property is part of and the new Mesh will be part of
            grid_uuid (UUID): the uuid of the RegularGrid object
            property_uuid (UUID): the uuid of a property whose supporting representation is a
                RegularGrid and where the indexable elements are columns

        returns:
            a new Mesh object with flavour reg&z, the xy regular points are the column centres and
            the z values are the property values

        notes:
            the grid's (constant) dx, dy & dz properties must be avaiable in the model;
            the property must not be a points property and must have a count of 1;
            discrete property data will be converted to continuous (float) data;
            if the property uom is of quantity class length, a suitable crs will be used for the mesh,
            otherwise the grid's crs will be used and a uom extra metadata item will be added;
            calling code will need to call the write_hdf5() and create_xml() methods for the mesh
        """
        grid = grr.RegularGrid(parent_model, uuid = grid_uuid)
        assert grid is not None
        assert grid.is_aligned
        prop = rqp.Property(parent_model, uuid = property_uuid, support_uuid = grid_uuid)
        assert prop is not None
        assert prop.indexable_element() == 'columns' and prop.count() == 1 and not prop.is_points()
        prop_array = prop.array_ref().astype(float)
        assert prop_array.shape == (grid.nj, grid.ni)
        crs = grid.crs
        em_uom = None
        if prop.is_continuous():
            uom = prop.uom()
            if uom != grid.crs.z_units and uom in wam.valid_uoms(quantity = 'length'):
                # create a copy of the grid's crs with the z units set to the property's uom
                crs = rqc.Crs(parent_model, uuid = grid.crs_uuid)
                crs.uuid = bu.new_uuid()
                crs.z_units = uom
                crs.create_xml(reuse = True)
            em_uom = uom
        else:  # use grid's crs and add extra metadata item
            em_uom = 'Euc'
        dxyz_dij = np.zeros((2, 3), dtype = float)
        dxyz_dij[0, 0] = grid.block_dxyz_dkji[2, 0]
        dxyz_dij[1, 1] = grid.block_dxyz_dkji[1, 1]
        assert dxyz_dij[0, 0] != 0.0 and dxyz_dij[1, 1] != 0.0
        em = prop.extra_metadata
        if em is None:
            em = {}
        if em_uom is not None:
            em['Uom'] = em_uom
        mesh = cls(parent_model,
                   mesh_flavour = 'reg&z',
                   nj = grid.nj,
                   ni = grid.ni,
                   origin = (0.5 * dxyz_dij[0, 0], 0.5 * dxyz_dij[1, 1], 0.0),
                   dxyz_dij = dxyz_dij,
                   z_values = prop_array,
                   surface_role = 'map',
                   crs_uuid = crs.uuid,
                   title = prop.title,
                   extra_metadata = em)
        return mesh

    def set_represented_interpretation_root(self, interp_root):
        """Makes a note of the xml root of the represented interpretation."""

        self.represented_interpretation_root = interp_root

    def full_array_ref(self):
        """Populates a full 2D(+1) numpy array of shape (nj, ni, 3) with xyz values, caches and returns.

        note:
           z values may be zero or not applicable when using the mesh as support for properties.
        """

        if self.full_array is not None:
            return self.full_array

        if self.flavour == 'explicit':
            self.__full_array_ref_explicit()
        elif self.flavour in ['regular', 'reg&z']:
            self.__full_array_ref_regular()
        elif self.flavour == 'ref&z':
            self.__full_array_ref_refz()
        else:
            raise Exception(f'unrecognised mesh flavour when fetching full array: {self.flavour}')

        return self.full_array

    def surface(self, quad_triangles = False):
        """Returns a surface object generated from this mesh."""

        return rqss.Surface(self.model, crs_uuid = self.crs_uuid, mesh = self, quad_triangles = quad_triangles)

    def write_hdf5(self, file_name = None, mode = 'a', use_xy_only = False):
        """Create or append to an hdf5 file, writing datasets for the mesh depending on flavour."""

        if not file_name:
            file_name = self.model.h5_file_name()
        if self.uuid is None:
            self.uuid = bu.new_uuid()
        if self.flavour == 'regular':
            return
        # NB: arrays must have been set up prior to calling this function
        h5_reg = rwh5.H5Register(self.model)
        a = self.full_array_ref()
        if self.flavour == 'explicit':
            if use_xy_only:
                h5_reg.register_dataset(self.uuid, 'points', a[..., :2])  # todo: check what others use here
            else:
                h5_reg.register_dataset(self.uuid, 'points', a)
        elif self.flavour == 'ref&z' or self.flavour == 'reg&z':
            h5_reg.register_dataset(self.uuid, 'zvalues', a[..., 2])
        else:
            log.error('bad mesh flavour when writing hdf5 array')
        h5_reg.write(file_name, mode = mode)

    def create_xml(self,
                   ext_uuid = None,
                   use_xy_only = False,
                   add_as_part = True,
                   add_relationships = True,
                   title = None,
                   originator = None):
        """Creates a grid 2d representation xml node from this mesh object and optionally adds as part of model.

        arguments:
           ext_uuid (uuid.UUID, optional): the uuid of the hdf5 external part holding the mesh array
           use_xy_only (boolean, default False): if True and the flavour of this mesh is explicit, only
              the xy coordinates are stored in the hdf5 dataset, otherwise xyz are stored
           add_as_part (boolean, default True): if True, the newly created xml node is added as a part
              in the model
           add_relationships (boolean, default True): if True, a relationship xml part is created relating the
              new grid 2d part to the crs part (and optional interpretation part), and to the represented
              interpretation if present
           title (string, optional): used as the citation Title text; should be meaningful to a human
           originator (string, optional): the name of the human being who created the grid 2d representation part;
              default is to use the login name

        returns:
           the newly created grid 2d representation (mesh) xml node
        """

        if ext_uuid is None and self.flavour != 'regular':
            ext_uuid = self.model.h5_uuid()

        g2d_node = super().create_xml(add_as_part = False, title = title, originator = originator)

        p_node = self.__create_xml_basics(g2d_node)

        ref_root = None

        if self.flavour == 'regular':
            self.__create_xml_regular(p_node)

        elif self.flavour == 'ref&z':
            self.__create_xml_refandz(p_node, ext_uuid)

        elif self.flavour == 'reg&z':
            self.__create_xml_regandz(p_node, ext_uuid)

        elif self.flavour == 'explicit':
            self.__create_xml_explicit(p_node, ext_uuid, use_xy_only)

        else:
            log.error('mesh has bad flavour when creating xml')
            return None

        if add_as_part:
            self.__create_xml_add_parts(g2d_node, ref_root, add_relationships, ext_uuid)

        return g2d_node

    def __load_from_arguments(self, ni, nj, origin, dxyz_dij, z_values, z_supporting_mesh_uuid, crs_uuid):
        assert ni is not None and nj is not None, 'If loading from arguments ni and nj must be provided'

        if z_values is None:
            assert origin is not None and dxyz_dij is not None and crs_uuid is not None, 'origin, dxyz_dij and crs_uuid must be provided for regular mesh'
            self.__regular_mesh_from_arguments(ni, nj, origin, dxyz_dij, z_values)
        else:
            if z_supporting_mesh_uuid is not None:
                self.__refz_from_arguments(ni, nj, z_values, z_supporting_mesh_uuid)
            else:
                assert dxyz_dij is not None and crs_uuid is not None and origin is not None, 'dzy_dij, crs_uuid and origin must be provided for regular and z mesh'
                self.__regz_from_arguments(ni, nj, z_values, origin, dxyz_dij)

    def __regz_from_arguments(self, ni, nj, z_values, origin, dxyz_dij):
        # create a reg&z mesh from arguments
        assert nj > 0 and ni > 0
        assert len(origin) == 3
        assert dxyz_dij.shape == (2, 3)
        assert z_values.shape == (nj, ni) or z_values.shape == (nj * ni,)
        self.nj = nj
        self.ni = ni
        self.regular_origin = origin
        self.regular_dxyz_dij = np.array(dxyz_dij, dtype = float)
        assert self.crs_uuid is not None, 'crs uuid missing'
        self.flavour = 'regular'
        self.full_array = None
        _ = self.full_array_ref()
        self.full_array[..., 2] = z_values.reshape(tuple(self.full_array.shape[:-1]))
        self.flavour = 'reg&z'

    def __refz_from_arguments(self, ni, nj, z_values, z_supporting_mesh_uuid):
        # create a ref&z mesh from arguments
        assert nj > 0 and ni > 0
        assert z_values.shape == (nj, ni) or z_values.shape == (nj * ni,)
        self.flavour = 'ref&z'
        self.nj = nj
        self.ni = ni
        self.ref_uuid = z_supporting_mesh_uuid
        self.ref_mesh = Mesh(self.model, uuid = z_supporting_mesh_uuid)
        assert self.ref_mesh is not None
        assert self.ref_mesh.nj == nj and self.ref_mesh.ni == ni
        self.full_array = self.ref_mesh.full_array_ref().copy()
        self.full_array[..., 2] = z_values.reshape(tuple(self.full_array.shape[:-1]))
        assert self.crs_uuid is not None, 'crs uuid missing'

    def __regular_mesh_from_arguments(self, ni, nj, origin, dxyz_dij, z_values):
        # create a regular mesh from arguments
        assert nj > 0 and ni > 0
        assert len(origin) == 3
        assert dxyz_dij.shape == (2, 3)
        self.flavour = 'regular'
        self.nj = nj
        self.ni = ni
        self.regular_origin = origin
        self.regular_dxyz_dij = np.array(dxyz_dij, dtype = float)
        assert self.crs_uuid is not None, 'crs uuid missing'

    def __load_from_xyz_values(self, xyz_values):
        # create an explicit mesh directly from a numpy array of points
        assert xyz_values.ndim == 3 and xyz_values.shape[2] == 3 and xyz_values.shape[0] > 1 and xyz_values.shape[1] > 1
        self.flavour = 'explicit'
        self.nj = xyz_values.shape[0]
        self.ni = xyz_values.shape[1]
        self.full_array = xyz_values.copy()
        assert self.crs_uuid is not None, 'crs uuid missing'

    def __load_from_mesh_file(self, mesh_file, mesh_flavour, mesh_format, crs_uuid, ni, nj):
        # load an explicit mesh from an ascii file in RMS text or zmap format
        assert mesh_format in ['rms', 'roxar', 'zmap']  # 'roxar' is treated synonymously with 'rms'
        assert mesh_flavour in ['explicit', 'regular', 'reg&z', 'ref&z']
        x, y, z = read_mesh(mesh_file, format = mesh_format)
        self.flavour = mesh_flavour
        self.nj = z.shape[0]
        self.ni = z.shape[1]
        assert self.nj > 1 and self.ni > 1
        self.full_array = np.stack((x, y, z), axis = -1)
        if mesh_flavour != 'explicit':
            min_x = x.flatten()[0]
            max_x = x.flatten()[-1]
            min_y = y.flatten()[0]
            max_y = y.flatten()[-1]
            dxyz_dij = np.array([[(max_x - min_x) /
                                  (self.ni - 1), 0.0, 0.0], [0.0, (max_y - min_y) / (self.nj - 1), 0.0]],
                                dtype = float)
            if mesh_flavour in ['regular', 'reg&z']:
                self.regular_origin = (min_x, min_y, 0.0)
                self.regular_dxyz_dij = dxyz_dij
            elif mesh_flavour == 'ref&z':
                self.ref_mesh = Mesh(self.model,
                                     ni = self.ni,
                                     nj = self.nj,
                                     origin = self.regular_origin,
                                     dxyz_dij = dxyz_dij,
                                     crs_uuid = crs_uuid)
                assert self.ref_mesh is not None
                assert self.ref_mesh.nj == nj and self.ref_mesh.ni == ni
            else:
                log.critical('code failure')
        assert self.crs_uuid is not None, 'crs uuid missing'
        # todo: option to create a regular and ref&z pair instead of an explicit mesh

    def __load_from_xml_regular(self, point_node):
        self.flavour = 'regular'
        origin_node = rqet.find_tag(point_node, 'Origin')
        assert origin_node is not None, 'origin missing in xml for regular mesh (lattice)'
        # yapf: disable
        self.regular_origin = (rqet.find_tag_float(origin_node, 'Coordinate1'),
                               rqet.find_tag_float(origin_node, 'Coordinate2'),
                               rqet.find_tag_float(origin_node, 'Coordinate3'))
        # yapf: enable
        offset_nodes = rqet.list_of_tag(point_node, 'Offset')  # first occurrence for FastestAxis, ie. I; 2nd for J
        assert len(offset_nodes) == 2, 'missing (or too many) offset nodes in xml for regular mesh (lattice)'
        self.regular_dxyz_dij = np.empty((2, 3))
        for j_or_i in range(2):  # 0 = J, 1 = I
            axial_offset_node = rqet.find_tag(offset_nodes[j_or_i], 'Offset')
            assert axial_offset_node is not None, 'missing offset offset node in xml'
            self.regular_dxyz_dij[1 - j_or_i] = (rqet.find_tag_float(axial_offset_node, 'Coordinate1'),
                                                 rqet.find_tag_float(axial_offset_node, 'Coordinate2'),
                                                 rqet.find_tag_float(axial_offset_node, 'Coordinate3'))
            if not maths.isclose(vec.dot_product(self.regular_dxyz_dij[1 - j_or_i], self.regular_dxyz_dij[1 - j_or_i]),
                                 1.0):
                log.warning('non-orthogonal axes and/or scaling in xml for regular mesh (lattice)')
            spacing_node = rqet.find_tag(offset_nodes[j_or_i], 'Spacing')
            stride = rqet.find_tag_float(spacing_node, 'Value')
            count = rqet.find_tag_int(spacing_node, 'Count')
            assert stride is not None and count is not None, 'missing spacing info in xml'
            assert count == (self.nj, self.ni)[j_or_i] - 1, \
                'unexpected value for count in xml spacing info for regular mesh (lattice)'
            assert stride > 0.0, 'spacing distance is not positive in xml for regular mesh (lattice)'
            self.regular_dxyz_dij[1 - j_or_i] *= stride

    def __load_from_xml_explicit(self, point_node):
        self.flavour = 'explicit'
        self.explicit_h5_key_pair = self.model.h5_uuid_and_path_for_node(point_node, tag = 'Coordinates')
        # load full_array on demand later (see full_array_ref() method)

    def __load_from_xml_refz(self, support_geom_node):
        self.flavour = 'ref&z'
        # assert rqet.node_type(support_geom_node) == 'Point3dFromRepresentationLatticeArray'  # only this supported for now
        self.ref_uuid = rqet.find_nested_tags_text(support_geom_node, ['SupportingRepresentation', 'UUID'])
        assert self.ref_uuid, 'missing supporting representation info in xml for z-value mesh'
        self.ref_mesh = Mesh(self.model, uuid = self.ref_uuid)
        assert self.nj == self.ref_mesh.nj and self.ni == self.ref_mesh.ni  # only this supported for now
        niosr_node = rqet.find_tag(support_geom_node, 'NodeIndicesOnSupportingRepresentation')
        start_value = rqet.find_tag_int(niosr_node, 'StartValue')
        assert start_value == 0, 'only full use of supporting mesh catered for at present'
        offset_nodes = rqet.list_of_tag(niosr_node, 'Offset')  # first occurrence for FastestAxis, ie. I; 2nd for J
        assert len(offset_nodes) == 2, 'missing (or too many) offset nodes in xml for regular mesh (lattice)'
        for j_or_i in range(2):  # 0 = J, 1 = I
            assert rqet.node_type(offset_nodes[j_or_i]) == 'IntegerConstantArray', 'variable step not catered for'
            assert rqet.find_tag_int(offset_nodes[j_or_i], 'Value') == 1, 'step other than 1 not catered for'
            count = rqet.find_tag_int(offset_nodes[j_or_i], 'Count')
            assert count == (self.nj, self.ni)[j_or_i] - 1, \
                'unexpected value for count in xml spacing info for regular mesh (lattice)'

    def __load_from_xml_regz(self, support_geom_node):
        self.flavour = 'reg&z'
        orig_node = rqet.find_tag(support_geom_node, 'Origin')
        self.regular_origin = (rqet.find_tag_float(orig_node,
                                                   'Coordinate1'), rqet.find_tag_float(orig_node, 'Coordinate2'),
                               rqet.find_tag_float(orig_node, 'Coordinate3'))
        assert self.regular_origin is not None, 'origin missing in xml for reg&z mesh (lattice)'
        offset_nodes = rqet.list_of_tag(support_geom_node,
                                        'Offset')  # first occurrence for FastestAxis, ie. I; 2nd for J
        assert len(offset_nodes) == 2, 'missing (or too many) offset nodes in xml for reg&z mesh (lattice)'
        self.regular_dxyz_dij = np.empty((2, 3))
        for j_or_i in range(2):  # 0 = J, 1 = I
            axial_offset_node = rqet.find_tag(offset_nodes[j_or_i], 'Offset')
            assert axial_offset_node is not None, 'missing offset offset node in xml'
            self.regular_dxyz_dij[1 - j_or_i] = (rqet.find_tag_float(axial_offset_node, 'Coordinate1'),
                                                 rqet.find_tag_float(axial_offset_node, 'Coordinate2'),
                                                 rqet.find_tag_float(axial_offset_node, 'Coordinate3'))
            if not maths.isclose(vec.dot_product(self.regular_dxyz_dij[1 - j_or_i], self.regular_dxyz_dij[1 - j_or_i]),
                                 1.0):
                log.warning('non-orthogonal axes and/or scaling in xml for regular mesh (lattice)')
            spacing_node = rqet.find_tag(offset_nodes[j_or_i], 'Spacing')
            stride = rqet.find_tag_float(spacing_node, 'Value')
            count = rqet.find_tag_int(spacing_node, 'Count')
            assert stride is not None and count is not None, 'missing spacing info in xml'
            assert count == (self.nj, self.ni)[j_or_i] - 1, \
                'unexpected value for count in xml spacing info for regular mesh (lattice)'
            assert stride > 0.0, 'spacing distance is not positive in xml for regular mesh (lattice)'
            self.regular_dxyz_dij[1 - j_or_i] *= stride

    def __load_from_xml_basics(self):
        root_node = self.root
        assert root_node is not None
        self.surface_role = rqet.find_tag_text(root_node, 'SurfaceRole')
        ref_node = rqet.find_tag(root_node, 'RepresentedInterpretation')
        if ref_node is not None:
            self.represented_interpretation_root = self.model.referenced_node(ref_node)
        patch_node = rqet.find_tag(root_node, 'Grid2dPatch')
        assert rqet.find_tag_int(patch_node, 'PatchIndex') == 0
        self.ni = rqet.find_tag_int(patch_node, 'FastestAxisCount')
        self.nj = rqet.find_tag_int(patch_node, 'SlowestAxisCount')
        assert self.ni is not None and self.nj is not None, 'mesh extent info missing in xml'
        geom_node = rqet.find_tag(patch_node, 'Geometry')
        assert geom_node is not None, 'geometry missing in mesh xml'
        self.crs_uuid = rqet.find_nested_tags_text(geom_node, ['LocalCrs', 'UUID'])
        assert self.crs_uuid is not None, 'crs reference missing in mesh geometry xml'
        point_node = rqet.find_tag(geom_node, 'Points')
        assert point_node is not None, 'missing Points node in mesh geometry xml'
        flavour = rqet.node_type(point_node)

        return point_node, flavour

    def _load_from_xml(self):
        point_node, flavour = self.__load_from_xml_basics()

        if flavour == 'Point3dLatticeArray':
            self.__load_from_xml_regular(point_node)
        elif flavour == 'Point3dZValueArray':
            # note: only simple, full use of supporting mesh is handled at present
            z_ref_node = rqet.find_tag(point_node, 'ZValues')
            self.ref_z_h5_key_pair = self.model.h5_uuid_and_path_for_node(z_ref_node, tag = 'Values')
            support_geom_node = rqet.find_tag(point_node, 'SupportingGeometry')
            if rqet.node_type(support_geom_node) == 'Point3dFromRepresentationLatticeArray':
                self.__load_from_xml_refz(support_geom_node)
            elif rqet.node_type(support_geom_node) == 'Point3dLatticeArray':
                self.__load_from_xml_regz(support_geom_node)
        elif flavour in ['Point3dHdf5Array', 'Point2dHdf5Array']:
            self.__load_from_xml_explicit(point_node)
        else:
            raise Exception('unrecognised flavour for mesh points')

    def __full_array_ref_explicit(self):
        # load array directly from hdf5 points reference; note: could be xyz or xy data
        assert self.explicit_h5_key_pair is not None, 'h5 key pair not established for mesh'
        try:
            self.model.h5_array_element(self.explicit_h5_key_pair,
                                        cache_array = True,
                                        object = self,
                                        array_attribute = 'full_array',
                                        dtype = 'float')
            # todo: could extend with z values of zero if only xy present?
        except Exception:
            log.exception('hdf5 points failure for mesh')

    def __full_array_ref_regular(self):
        self.full_array = np.empty((self.nj, self.ni, 3))
        x_i0 = np.linspace(0.0, (self.nj - 1) * self.regular_dxyz_dij[1, 0], num = self.nj)
        y_i0 = np.linspace(0.0, (self.nj - 1) * self.regular_dxyz_dij[1, 1], num = self.nj)
        z_i0 = np.linspace(0.0, (self.nj - 1) * self.regular_dxyz_dij[1, 2], num = self.nj)
        x_full = np.linspace(x_i0, x_i0 + (self.ni - 1) * self.regular_dxyz_dij[0, 0], num = self.ni, axis = -1)
        y_full = np.linspace(y_i0, y_i0 + (self.ni - 1) * self.regular_dxyz_dij[0, 1], num = self.ni, axis = -1)
        z_full = np.linspace(z_i0, z_i0 + (self.ni - 1) * self.regular_dxyz_dij[0, 2], num = self.ni, axis = -1)
        self.full_array = np.stack((x_full, y_full, z_full), axis = -1) + self.regular_origin

        if self.flavour == 'reg&z':  # overwrite regular z values with explicitly stored z values
            self.__full_array_ref_regz()

    def __full_array_ref_regz(self):
        assert self.ref_z_h5_key_pair is not None, 'h5 key pair missing for mesh z values'
        try:
            self.model.h5_array_element(self.ref_z_h5_key_pair,
                                        cache_array = True,
                                        object = self,
                                        array_attribute = 'temp_z',
                                        dtype = 'float')
        except Exception:
            log.exception('hdf5 failure for mesh z values')
        self.full_array[..., 2] = self.temp_z
        delattr(self, 'temp_z')

    def __full_array_ref_refz(self):
        # load array from referenced mesh and overwrite z values
        if self.ref_mesh is None:
            self.ref_mesh = Mesh(self.model, uuid = self.ref_uuid)
            assert self.ref_mesh is not None, 'failed to instantiate object for referenced mesh'
        self.full_array = self.ref_mesh.full_array_ref().copy()
        assert self.ref_z_h5_key_pair is not None, 'h5 key pair missing for mesh z values'
        try:
            self.model.h5_array_element(self.ref_z_h5_key_pair,
                                        cache_array = True,
                                        object = self,
                                        array_attribute = 'temp_z',
                                        dtype = 'float')
        except Exception:
            log.exception('hdf5 failure for mesh z values')
        self.full_array[..., 2] = self.temp_z
        #        if np.any(np.isnan(self.temp_z)): log.warning('z values include some NaNs')
        delattr(self, 'temp_z')

    def __create_xml_regular(self, p_node):

        assert self.regular_origin is not None and self.regular_dxyz_dij is not None

        p_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dLatticeArray')

        self.model.create_solitary_point3d('Origin', p_node, self.regular_origin)  # todo: check xml namespace

        for j_or_i in range(2):  # 0 = J, 1 = I
            dxyz = self.regular_dxyz_dij[1 - j_or_i].copy()
            log.debug('dxyz: ' + str(dxyz))
            d_value = vec.dot_product(dxyz, dxyz)
            assert d_value > 0.0
            d_value = maths.sqrt(d_value)
            dxyz /= d_value
            o_node = rqet.SubElement(p_node, ns['resqml2'] + 'Offset')  # note: 1st Offset is for J axis , 2nd for I
            o_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dOffset')
            o_node.text = '\n'
            self.model.create_solitary_point3d('Offset', o_node, dxyz)
            space_node = rqet.SubElement(o_node, ns['resqml2'] + 'Spacing')
            space_node.set(ns['xsi'] + 'type',
                           ns['resqml2'] + 'DoubleConstantArray')  # nothing else catered for just now
            space_node.text = '\n'
            ov_node = rqet.SubElement(space_node, ns['resqml2'] + 'Value')
            ov_node.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
            ov_node.text = str(d_value)
            oc_node = rqet.SubElement(space_node, ns['resqml2'] + 'Count')
            oc_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            if j_or_i:
                oc_node.text = str(self.ni - 1)
            else:
                oc_node.text = str(self.nj - 1)

    def __create_xml_refandz(self, p_node, ext_uuid):
        assert ext_uuid is not None
        assert self.ref_uuid is not None

        p_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dZValueArray')

        sg_node = rqet.SubElement(p_node, ns['resqml2'] + 'SupportingGeometry')

        sg_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dFromRepresentationLatticeArray')
        sg_node.text = '\n'

        niosr_node = rqet.SubElement(sg_node, ns['resqml2'] + 'NodeIndicesOnSupportingRepresentation')
        niosr_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerLatticeArray')
        niosr_node.text = '\n'

        sv_node = rqet.SubElement(niosr_node, ns['resqml2'] + 'StartValue')
        sv_node.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
        sv_node.text = '0'  # no other possibility cater for at present

        for j_or_i in range(2):  # 0 = J, 1 = I
            o_node = rqet.SubElement(niosr_node, ns['resqml2'] + 'Offset')
            o_node.set(ns['xsi'] + 'type',
                       ns['resqml2'] + 'IntegerConstantArray')  # no other possibility cater for at present
            o_node.text = '\n'
            ov_node = rqet.SubElement(o_node, ns['resqml2'] + 'Value')
            ov_node.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
            ov_node.text = '1'  # no other possibility cater for at present
            oc_node = rqet.SubElement(o_node, ns['resqml2'] + 'Count')
            oc_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            if j_or_i:
                oc_node.text = str(self.ni - 1)
            else:
                oc_node.text = str(self.nj - 1)

        ref_root = self.model.root_for_uuid(self.ref_uuid)
        self.model.create_ref_node('SupportingRepresentation',
                                   rqet.find_nested_tags_text(ref_root, ['Citation', 'Title']),
                                   self.ref_uuid,
                                   content_type = 'Grid2dRepresentation',
                                   root = sg_node)

        zv_node = rqet.SubElement(p_node, ns['resqml2'] + 'ZValues')
        zv_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'DoubleHdf5Array')
        zv_node.text = '\n'

        v_node = rqet.SubElement(zv_node, ns['resqml2'] + 'Values')
        v_node.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        v_node.text = '\n'

        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'zvalues', root = v_node)

    def __create_xml_regandz(self, p_node, ext_uuid):

        assert ext_uuid is not None

        p_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dZValueArray')

        sg_node = rqet.SubElement(p_node, ns['resqml2'] + 'SupportingGeometry')

        sg_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dLatticeArray')
        sg_node.text = '\n'

        assert self.regular_origin is not None and self.regular_dxyz_dij is not None

        self.model.create_solitary_point3d('Origin', sg_node, self.regular_origin)  # todo: check xml namespace

        for j_or_i in range(2):  # 0 = J, 1 = I; ie. J axis info first in xml, followed by I axis
            dxyz = self.regular_dxyz_dij[1 - j_or_i].copy()
            log.debug('dxyz: ' + str(dxyz))
            d_value = vec.dot_product(dxyz, dxyz)
            assert d_value > 0.0
            d_value = maths.sqrt(d_value)
            dxyz /= d_value
            o_node = rqet.SubElement(sg_node, ns['resqml2'] + 'Offset')
            o_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dOffset')
            o_node.text = '\n'
            self.model.create_solitary_point3d('Offset', o_node, dxyz)
            space_node = rqet.SubElement(o_node, ns['resqml2'] + 'Spacing')
            space_node.set(ns['xsi'] + 'type',
                           ns['resqml2'] + 'DoubleConstantArray')  # nothing else catered for just now
            space_node.text = '\n'
            ov_node = rqet.SubElement(space_node, ns['resqml2'] + 'Value')
            ov_node.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
            ov_node.text = str(d_value)
            oc_node = rqet.SubElement(space_node, ns['resqml2'] + 'Count')
            oc_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            if j_or_i:
                oc_node.text = str(self.ni - 1)
            else:
                oc_node.text = str(self.nj - 1)

        zv_node = rqet.SubElement(p_node, ns['resqml2'] + 'ZValues')
        zv_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'DoubleHdf5Array')
        zv_node.text = '\n'

        v_node = rqet.SubElement(zv_node, ns['resqml2'] + 'Values')
        v_node.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        v_node.text = '\n'

        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'zvalues', root = v_node)

    def __create_xml_explicit(self, p_node, ext_uuid, use_xy_only):
        assert ext_uuid is not None

        if use_xy_only:
            p_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point2dHdf5Array')
        else:
            p_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')

        coords = rqet.SubElement(p_node, ns['resqml2'] + 'Coordinates')
        coords.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        coords.text = '\n'

        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'points', root = coords)

    def __create_xml_basics(self, g2d_node):
        if self.represented_interpretation_root is not None:
            interp_root = self.represented_interpretation_root
            interp_uuid = bu.uuid_from_string(interp_root.attrib['uuid'])
            interp_part = self.model.part_for_uuid(interp_uuid)
            self.model.create_ref_node('RepresentedInterpretation',
                                       rqet.find_nested_tags_text(interp_root, ['Citation', 'Title']),
                                       interp_uuid,
                                       content_type = self.model.type_of_part(interp_part),
                                       root = g2d_node)

        role_node = rqet.SubElement(g2d_node, ns['resqml2'] + 'SurfaceRole')
        role_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'SurfaceRole')
        role_node.text = self.surface_role

        patch_node = rqet.SubElement(g2d_node, ns['resqml2'] + 'Grid2dPatch')
        patch_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Grid2dPatch')
        patch_node.text = '\n'

        pi_node = rqet.SubElement(patch_node, ns['resqml2'] + 'PatchIndex')
        pi_node.set(ns['xsi'] + 'type', ns['xsd'] + 'nonNegativeInteger')
        pi_node.text = '0'

        fast_node = rqet.SubElement(patch_node, ns['resqml2'] + 'FastestAxisCount')
        fast_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        fast_node.text = str(self.ni)

        slow_node = rqet.SubElement(patch_node, ns['resqml2'] + 'SlowestAxisCount')
        slow_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        slow_node.text = str(self.nj)

        geom = rqet.SubElement(patch_node, ns['resqml2'] + 'Geometry')
        geom.set(ns['xsi'] + 'type', ns['resqml2'] + 'PointGeometry')
        geom.text = '\n'

        self.model.create_crs_reference(crs_uuid = self.crs_uuid, root = geom)

        p_node = rqet.SubElement(geom, ns['resqml2'] + 'Points')
        p_node.text = '\n'

        return p_node

    def __create_xml_add_parts(self, g2d_node, ref_root, add_relationships, ext_uuid):
        self.model.add_part('obj_Grid2dRepresentation', self.uuid, g2d_node)
        if add_relationships:
            crs_root = self.model.root_for_uuid(self.crs_uuid)
            self.model.create_reciprocal_relationship(g2d_node, 'destinationObject', crs_root, 'sourceObject')
            if self.represented_interpretation_root is not None:
                self.model.create_reciprocal_relationship(g2d_node, 'destinationObject',
                                                          self.represented_interpretation_root, 'sourceObject')
            if ref_root is not None:  # used for ref&z flavour
                self.model.create_reciprocal_relationship(g2d_node, 'destinationObject', ref_root, 'sourceObject')
            if self.flavour == 'ref&z' or self.flavour == 'explicit' or self.flavour == 'reg&z':
                ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
                ext_node = self.model.root_for_part(ext_part)
                self.model.create_reciprocal_relationship(g2d_node, 'mlToExternalPartProxy', ext_node,
                                                          'externalPartProxyToMl')



"""Multiprocessing module containing the function used to run the wrapper functions in parallel."""

import logging
import os
import time
import uuid
import joblib  # type: ignore
from typing import List, Dict, Any, Callable, Union
from pathlib import Path
from joblib import Parallel, delayed, parallel_backend  # type: ignore

import resqpy.model as rq
import resqpy.olio.consolidation as cons

log = logging.getLogger(__name__)


def rm_tree(path: Union[Path, str]) -> None:
    """Removes a directory using a pathlib Path.

    arguments:
        path (Path or str): pathlib Path or string of the directory path
    """
    path = Path(path)
    if os.path.isdir(path):
        for child in path.iterdir():
            if child.is_file():
                child.unlink()
            else:
                rm_tree(child)
        path.rmdir()


def function_multiprocessing(function: Callable,
                             kwargs_list: List[Dict[str, Any]],
                             recombined_epc: Union[Path, str],
                             cluster,
                             consolidate: bool = True,
                             require_success = False,
                             tmp_dir_path: Union[Path, str] = '.',
                             backend: str = 'dask',
                             clean_up: bool = True) -> List[bool]:
    """Calls a function concurrently with the specfied arguments.

    arguments:
        function (Callable): the wrapper function to be called; needs to return:
            - index (int): the index of the kwargs in the kwargs_list;
            - success (bool): whether the function call was successful, however that is defined;
            - epc_file (Path or str): the epc file path where the objects are stored;
            - uuid_list (list of str): list of UUIDs of relevant objects;
        kwargs_list (list of dict): A list of keyword argument dictionaries that are
            used when calling the function
        recombined_epc (Path or str): A pathlib Path or path string of
            where the combined epc will be saved
        cluster: if using the Dask backend, a LocalCluster is a Dask cluster on a
            local machine. If using a job queing system, a JobQueueCluster can be used
            such as an SGECluster, SLURMCluster, PBSCluster, LSFCluster etc
        consolidate (bool): if True and an equivalent part already exists in
            a model, it is not duplicated and the uuids are noted as equivalent
        require_success (bool): if True and any instance fails, then an exception is
            raised
        tmp_dir_path (str): path where the temporary directory is saved; defaults to
            the calling code directory
        backend (str): the joblib parallel backend used. Dask is used by default
            so a Dask cluster must be passed to the cluster argument
        clean_up (bool, default True): if True, the temporary directory used during
            multi processing is deleted; if False, it is left in place with its
            contents (to facilitate debugging)

    returns:
        success_list (List[bool]): A boolean list of successful function calls

    notes:
        a multiprocessing pool is used to call the function multiple times in parallel;
        once all results are returned, they are combined into a single epc file;
        this function uses the Dask backend by default to run the given function in
        parallel, so a Dask cluster must be setup and passed as an argument if Dask is
        used; Dask will need to be installed in the Python environment because it is not
        a dependency of the project; more info can be found at
        https://resqpy.readthedocs.io/en/latest/tutorial/multiprocessing.html
    """
    log.info("multiprocessing function called with %s function, %s entries.", function.__name__, len(kwargs_list))

    if tmp_dir_path is None:
        tmp_dir_path = '.'
    tmp_dir = Path(tmp_dir_path) / f'tmp_{uuid.uuid4()}'
    os.makedirs(tmp_dir)
    for i, kwargs in enumerate(kwargs_list):
        kwargs["index"] = i
        kwargs["parent_tmp_dir"] = str(tmp_dir)

    if cluster is None:
        results = []
        for i, kwargs in enumerate(kwargs_list):
            # log.debug(f'calling function for entry {i}; name: {kwargs.get("name")}')
            one_r = function(**kwargs)
            results.append(one_r)
            # log.debug(f'completed entry: {one_r[0]}; success: {one_r[1]}; epc: {one_r[2]}')
            # log.debug(f'uuid list: {one_r[3]}')
    else:
        with parallel_backend(backend):
            results = Parallel()(delayed(function)(**kwargs) for kwargs in kwargs_list)

    # Sorting the results by the original kwargs_list index.
    results = list(sorted(results, key = lambda x: x[0]))

    success_list = [result[1] for result in results]
    epc_list = [result[2] for result in results]
    uuids_list = [result[3] for result in results]
    success_count = sum(success_list)
    log.info("multiprocessing function calls complete; successes: %s/%s.", success_count, len(results))
    if require_success and success_count < len(results):
        raise Exception('one or more multiprocessing instances failed')

    epc_file = Path(str(recombined_epc))
    if epc_file.is_file():
        model_recombined = rq.Model(epc_file = str(epc_file), quiet = True)
        log.info(f"updating the recombined epc file: {epc_file}")
    else:
        model_recombined = rq.new_model(epc_file = str(epc_file))
        log.info(f"creating the recombined epc file: {epc_file}")

    model = None
    for i, epc in enumerate(epc_list):
        # log.debug(f'recombining from mp instance {i} epc: {epc}')
        if epc is None:
            continue
        attempt = 0
        while not os.path.exists(epc):
            attempt += 1
            if attempt == 7:
                log.warning(f'mp epc slow to materialise: {epc}')
            if attempt > 300:
                raise FileNotFoundError(f'timeout waiting for multiprocess worker epc to become available: {epc}')
            time.sleep(min(attempt, 10))
        attempt = 0
        while True:
            attempt += 1
            try:
                model = rq.Model(epc_file = epc, quiet = True)
                break
            except FileNotFoundError:
                if attempt >= 10:
                    raise FileNotFoundError(f'timeout waiting for mp epc {epc}')
                time.sleep(1)
        uuids = cons.sort_uuids_list(model, uuids_list[i])
        if uuids is None:
            uuids = model.uuids()
        for u in uuids:
            attempt = 0
            while True:
                attempt += 1
                try:
                    model_recombined.copy_uuid_from_other_model(model, uuid = u, consolidate = consolidate)
                    break
                except BlockingIOError:
                    if attempt >= 5:
                        raise
                time.sleep(1)

    # Deleting temporary directory.
    # log.debug(f"deleting the temporary directory {tmp_dir}")
    if clean_up:
        rm_tree(tmp_dir)

    model_recombined.store_epc(quiet = True)
    if model is not None:
        model.h5_release()

    log.debug(f"recombined epc file complete: {epc_file}")

    return success_list
